import unittest

import weaviate


class TestVersion(unittest.TestCase):
    def test_version(self):
        """
        Test the `__version__` global variable.
        """

        self.assertEqual(weaviate.__version__, "3.15.2", "Check if the version is set correctly!")



"""
GraphQL `Get` command.
"""

from typing import List
from weaviate.gql.filter import (
    GraphQL,
)
from weaviate.connect import Connection
from .get import GetBuilder


class MultiGetBuilder(GraphQL):
    """
    GetBuilder class used to create GraphQL queries.
    """

    def __init__(self, get_builder: List[GetBuilder], connection: Connection):
        """
        Initialize a MultiGetBuilder class instance.

        Parameters
        ----------
        get_builder : list of GetBuilder
            GetBuilder objects for a single request each.
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.

        Examples
        --------
        To create a 'multi_get' object using several 'get' request at the same time use:

        >>>    client.query.multi_get(
        ... [
        ...    client.query.get("Ship", ["name"]).with_alias("one"),
        ...    client.query.get("Ship", ["size"]).with_alias("two"),
        ...    client.query.get("Person", ["name"])
        ... ]
        with_alias() needs to be used if the same 'class_name' is used twice during the same 'multi_get' request.

        Raises
        ------
        TypeError
            If 'get_builder' is of wrong type.
        """
        get_names = []
        super().__init__(connection)
        if not isinstance(get_builder, List):
            raise TypeError(f"get_builder must be of type List but was {type(get_builder)}")
        for get in get_builder:
            if not isinstance(get, GetBuilder):
                raise TypeError(
                    f"All objects in 'get_builder' must be of type 'GetBuilder' but at least one object was {type(get)}"
                )
            if get.name not in get_names:
                get_names.append(get.name)
            else:
                raise TypeError(
                    f"Objects in 'get_builder' can not have duplicate names but two were named: '{get.name}'. Queries can be renamed using an alias."
                )
        self.get_builder: List[GetBuilder] = get_builder

    def build(self) -> str:
        """
        Build query filter as a string.

        Returns
        -------
        str
            The GraphQL query as a string.
        """
        query = "{Get{"

        for get in self.get_builder:
            query += get.build(wrap_get=False)
        return query + "}}"



import hashlib
import os
import platform
import signal
import socket
import stat
import subprocess
import tarfile
import time
import urllib.request
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional

from weaviate.exceptions import WeaviateStartUpError

DEFAULT_BINARY_PATH = str((Path.home() / ".cache/weaviate-embedded/"))
DEFAULT_PERSISTENCE_DATA_PATH = str((Path.home() / ".local/share/weaviate"))
GITHUB_RELEASE_DOWNLOAD_URL = "https://github.com/weaviate/weaviate/releases/download/"


@dataclass
class EmbeddedOptions:
    persistence_data_path: str = os.environ.get("XDG_DATA_HOME", DEFAULT_PERSISTENCE_DATA_PATH)
    binary_path: str = os.environ.get("XDG_CACHE_HOME", DEFAULT_BINARY_PATH)
    version: str = "https://github.com/weaviate/weaviate/releases/download/v1.18.1/weaviate-v1.18.1-linux-amd64.tar.gz"
    port: int = 6666
    hostname: str = "127.0.0.1"
    cluster_hostname: str = "embedded"
    additional_env_vars: Optional[Dict[str, str]] = None


def get_random_port() -> int:
    sock = socket.socket()
    sock.bind(("", 0))
    port_num = int(sock.getsockname()[1])
    sock.close()
    return port_num


class EmbeddedDB:
    def __init__(self, options: EmbeddedOptions):
        self.data_bind_port = get_random_port()
        self.options = options
        self.pid = 0
        self.ensure_paths_exist()
        self.check_supported_platform()
        self._parsed_weaviate_version = ""
        if self.options.version.startswith(GITHUB_RELEASE_DOWNLOAD_URL):
            # replace with str.removeprefix() after 3.8 has been deprecated
            self._parsed_weaviate_version = self.options.version[
                len(GITHUB_RELEASE_DOWNLOAD_URL) :
            ].split("/")[0]

    def __del__(self):
        self.stop()

    def ensure_paths_exist(self):
        Path(self.options.binary_path).mkdir(parents=True, exist_ok=True)
        Path(self.options.persistence_data_path).mkdir(parents=True, exist_ok=True)

    def ensure_weaviate_binary_exists(self):
        self._weaviate_binary_path = Path(
            self.options.binary_path,
            "weaviate-"
            + self._parsed_weaviate_version
            + "-"
            + str(hashlib.sha256(self.options.version.encode("utf-8")).hexdigest()),
        )
        if not self._weaviate_binary_path.exists():
            print(
                f"Binary {self.options.binary_path} did not exist. Downloading binary from {self.options.version}"
            )
            tar_filename = Path(self.options.binary_path, "tmp_weaviate.tgz")
            urllib.request.urlretrieve(self.options.version, tar_filename)
            binary_tar = tarfile.open(tar_filename)
            binary_tar.extract("weaviate", path=Path(self.options.binary_path))
            (Path(self.options.binary_path) / "weaviate").rename(self._weaviate_binary_path)
            tar_filename.unlink()

            # Ensuring weaviate binary is executable
            self._weaviate_binary_path.chmod(
                self._weaviate_binary_path.stat().st_mode | stat.S_IEXEC
            )

    def is_listening(self) -> bool:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.options.hostname, self.options.port))
            s.close()
            return True
        except (socket.error, ConnectionRefusedError):
            s.close()
            return False

    def wait_till_listening(self):
        seconds = 30
        sleep_interval = 0.1
        retries = int(seconds / sleep_interval)
        while self.is_listening() is False and retries > 0:
            time.sleep(sleep_interval)
            retries -= 1
        if retries == 0:
            raise WeaviateStartUpError(
                f"Embedded DB did not start listening on port {self.options.port} within {seconds} seconds"
            )

    @staticmethod
    def check_supported_platform():
        if platform.system() in ["Darwin", "Windows"]:
            raise WeaviateStartUpError(
                f"{platform.system()} is not supported with EmbeddedDB. Please upvote the feature request if "
                f"you want this: https://github.com/weaviate/weaviate-python-client/issues/239"
            )

    def start(self):
        if self.is_listening():
            print(f"embedded weaviate is already listing on port {self.options.port}")
            return

        self.ensure_weaviate_binary_exists()
        my_env = os.environ.copy()

        my_env.setdefault("AUTHENTICATION_ANONYMOUS_ACCESS_ENABLED", "true")
        my_env.setdefault("QUERY_DEFAULTS_LIMIT", "20")
        my_env.setdefault("PERSISTENCE_DATA_PATH", self.options.persistence_data_path)
        my_env.setdefault("CLUSTER_HOSTNAME", self.options.cluster_hostname)
        # Bug with weaviate requires setting gossip and data bind port
        my_env.setdefault("CLUSTER_GOSSIP_BIND_PORT", str(get_random_port()))
        my_env.setdefault(
            "ENABLE_MODULES",
            "text2vec-openai,text2vec-cohere,text2vec-huggingface,ref2vec-centroid,generative-openai,qna-openai",
        )

        if self.options.additional_env_vars is not None:
            my_env.update(self.options.additional_env_vars)

        # filter warning about running processes.
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", ResourceWarning)
            process = subprocess.Popen(
                [
                    f"{self._weaviate_binary_path}",
                    "--host",
                    self.options.hostname,
                    "--port",
                    str(self.options.port),
                    "--scheme",
                    "http",
                ],
                env=my_env,
            )
            self.pid = process.pid
        print(f"Started {self.options.binary_path}: process ID {self.pid}")
        self.wait_till_listening()

    def stop(self):
        if self.pid > 0:
            try:
                os.kill(self.pid, signal.SIGTERM)
            except ProcessLookupError:
                print(
                    f"Tried to stop embedded weaviate process {self.pid}. Process {self.pid} "
                    f"was not found. So not doing anything"
                )

    def ensure_running(self):
        if not self.is_listening():
            print(
                f"Embedded weaviate wasn't listening on port {self.options.port}, so starting embedded weaviate again"
            )
            self.start()



"""
Weaviate Exceptions.
"""

from requests import Response, exceptions

ERROR_CODE_EXPLANATION = {
    413: """Payload Too Large. Try to decrease the batch size or increase the maximum request size on your weaviate
         server."""
}


class WeaviateBaseError(Exception):
    """
    Weaviate base exception that all Weaviate exceptions should inherit from.
    This error can be used to catch any Weaviate exceptions.
    """

    def __init__(self, message: str = ""):
        """
        Weaviate base exception initializer.
        Parameters
        ----------
        message: str, optional
            An error message specific to the context in which the error occurred.
        """

        self.message = message
        super().__init__(message)


class UnexpectedStatusCodeException(WeaviateBaseError):
    """
    Is raised in case the status code returned from Weaviate is
    not handled in the client implementation and suggests an error.
    """

    def __init__(self, message: str, response: Response):
        """
        Is raised in case the status code returned from Weaviate is
        not handled in the client implementation and suggests an error.

        Custom code can act on the attributes:
        - status_code
        - json

        Parameters
        ----------
        message: str
            An error message specific to the context, in which the error occurred.
        response: requests.Response
            The request response of which the status code was unexpected.
        """
        self._status_code: int = response.status_code
        # Set error message

        try:
            body = response.json()
        except exceptions.JSONDecodeError:
            body = None

        msg = (
            message
            + f"! Unexpected status code: {response.status_code}, with response body: {body}."
        )
        if response.status_code in ERROR_CODE_EXPLANATION:
            msg += " " + ERROR_CODE_EXPLANATION[response.status_code]

        super().__init__(msg)

    @property
    def status_code(self) -> int:
        return self._status_code


class ObjectAlreadyExistsException(WeaviateBaseError):
    """
    Object Already Exists Exception.
    """


class AuthenticationFailedException(WeaviateBaseError):
    """
    Authentication Failed Exception.
    """


class SchemaValidationException(WeaviateBaseError):
    """
    Schema Validation Exception.
    """


class BackupFailedException(WeaviateBaseError):
    """
    Backup Failed Exception.
    """


class EmptyResponseException(WeaviateBaseError):
    """
    Occurs when an HTTP request unexpectedly returns an empty response
    """


class MissingScopeException(WeaviateBaseError):
    """Scope was not provided with client credential flow."""


class WeaviateStartUpError(WeaviateBaseError):
    """Is raised if weaviate does not start up in time."""



from __future__ import annotations

from typing import Dict, List, Union
from typing import TYPE_CHECKING

import requests
from authlib.integrations.requests_client import OAuth2Session

from weaviate.auth import (
    AuthCredentials,
    AuthClientPassword,
    AuthBearerToken,
    AuthClientCredentials,
)
from weaviate.exceptions import MissingScopeException, AuthenticationFailedException
from ..warnings import _Warnings

if TYPE_CHECKING:
    from . import BaseConnection

AUTH_DEFAULT_TIMEOUT = 5


class _Auth:
    def __init__(
        self,
        oidc_config: Dict[str, Union[str, List[str]]],
        credentials: AuthCredentials,
        connection: BaseConnection,
    ) -> None:
        self._credentials: AuthCredentials = credentials
        self._connection: BaseConnection = connection

        self._open_id_config_url: str = oidc_config["href"]
        self._client_id: str = oidc_config["clientId"]
        self._default_scopes: List[str] = []
        if "scopes" in oidc_config:
            self._default_scopes = oidc_config["scopes"]

        self._token_endpoint: str = self._get_token_endpoint()
        self._validate(oidc_config)

    def _validate(self, oidc_config: Dict[str, str]) -> None:
        if isinstance(self._credentials, AuthClientPassword):
            if self._token_endpoint.startswith("https://login.microsoftonline.com"):
                raise AuthenticationFailedException(
                    """Microsoft/azure does not recommend to authenticate using username and password and this method is
                    not supported by the python client."""
                )

            # The grant_types_supported field is optional and does not have to be present in the response
            if (
                "grant_types_supported" in oidc_config
                and "password" not in oidc_config["grant_types_supported"]
            ):
                raise AuthenticationFailedException(
                    """The grant_types supported by the third-party authentication service are insufficient. Please add
                    the 'password' grant type."""
                )

    def _get_token_endpoint(self) -> str:
        response_auth = requests.get(self._open_id_config_url, proxies=self._connection.proxies)
        return response_auth.json()["token_endpoint"]

    def get_auth_session(self) -> OAuth2Session:
        if isinstance(self._credentials, AuthBearerToken):
            session = self._get_session_auth_bearer_token(self._credentials)
        elif isinstance(self._credentials, AuthClientCredentials):
            session = self._get_session_client_credential(self._credentials)
        else:
            assert isinstance(self._credentials, AuthClientPassword)
            session = self._get_session_user_pw(self._credentials)

        return session

    def _get_session_auth_bearer_token(self, config: AuthBearerToken) -> OAuth2Session:
        token = {"access_token": config.access_token}
        if config.expires_in is not None:
            token["expires_in"] = config.expires_in
        if config.refresh_token is not None:
            token["refresh_token"] = config.refresh_token

        if "refresh_token" not in token:
            _Warnings.auth_no_refresh_token(config.expires_in)

        # token endpoint and clientId are needed for token refresh
        return OAuth2Session(
            token=token,
            token_endpoint=self._token_endpoint,
            client_id=self._client_id,
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )

    def _get_session_user_pw(self, config: AuthClientPassword) -> OAuth2Session:
        scope: List[str] = self._default_scopes.copy()
        scope.extend(config.scope)
        session = OAuth2Session(
            client_id=self._client_id,
            token_endpoint=self._token_endpoint,
            grant_type="password",
            scope=scope,
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )
        token = session.fetch_token(username=config.username, password=config.password)
        if "refresh_token" not in token:
            _Warnings.auth_no_refresh_token(token["expires_in"])

        return session

    def _get_session_client_credential(self, config: AuthClientCredentials) -> OAuth2Session:
        scope: List[str] = self._default_scopes.copy()

        if config.scope is not None:
            scope.extend(config.scope)
        if len(scope) == 0:
            # hardcode commonly used scopes
            if self._token_endpoint.startswith("https://login.microsoftonline.com"):
                scope = [self._client_id + "/.default"]
            else:
                raise MissingScopeException

        session = OAuth2Session(
            client_id=self._client_id,
            client_secret=config.client_secret,
            token_endpoint_auth_method="client_secret_post",
            scope=scope,
            token_endpoint=self._token_endpoint,
            grant_type="client_credentials",
            token={"access_token": None, "expires_in": -100},
            default_timeout=AUTH_DEFAULT_TIMEOUT,
        )
        # explicitly fetch tokens. Otherwise, authlib will do it in the background and we might have race-conditions
        session.fetch_token()
        return session



import os
import signal
import socket
import tarfile
import time
import uuid
from pathlib import Path
from sys import platform
from unittest.mock import patch

import pytest
from pytest_httpserver import HTTPServer
from werkzeug import Request, Response

import weaviate
from weaviate import embedded
from weaviate.embedded import EmbeddedDB, EmbeddedOptions
from weaviate.exceptions import WeaviateStartUpError

if platform != "linux":
    pytest.skip("Currently only supported on linux", allow_module_level=True)


def test_embedded__init__(tmp_path):
    assert (
        EmbeddedDB(EmbeddedOptions(port=6666, persistence_data_path=tmp_path)).options.port == 6666
    )


def test_embedded__init__non_default_port(tmp_path):
    assert (
        EmbeddedDB(EmbeddedOptions(port=30666, persistence_data_path=tmp_path)).options.port
        == 30666
    )


def test_embedded_ensure_binary_exists(tmp_path):
    bin_path = tmp_path / "notcreated-yet/bin/weaviate-embedded"
    assert bin_path.is_file, False
    embedded_db = EmbeddedDB(
        EmbeddedOptions(binary_path=str(bin_path), persistence_data_path=tmp_path / "2")
    )
    embedded_db.ensure_weaviate_binary_exists()
    assert Path(embedded_db.options.binary_path).is_file, True


def test_version_parsing(tmp_path):
    bin_path = tmp_path / "bin"
    embedded_db = EmbeddedDB(
        EmbeddedOptions(
            binary_path=str(bin_path),
            version="https://github.com/weaviate/weaviate/releases/download/v1.18.1/weaviate-v1.18.1-linux-amd64.tar.gz",
            persistence_data_path=tmp_path / "2",
        )
    )
    embedded_db.ensure_weaviate_binary_exists()
    embedded_file_name = list(bin_path.iterdir())
    assert len(embedded_file_name) == 1  # .tgz file was deleted
    assert "v1.18.1" in str(embedded_file_name[0])


def test_download_no_version_parsing(httpserver: HTTPServer, tmp_path):
    """Test downloading weaviate from a non-github url."""

    def handler(request: Request):
        with open(Path(tmp_path, "weaviate"), "w") as _:
            with tarfile.open(Path(tmp_path, "tmp_weaviate.tgz"), "w:gz") as tar:
                tar.add(Path(tmp_path, "weaviate"), arcname="weaviate")

        return Response(open(Path(tmp_path, "tmp_weaviate.tgz"), mode="rb"))

    httpserver.expect_request("/tmp_weaviate.tgz").respond_with_handler(handler)

    bin_path = tmp_path / "bin"
    embedded_db = EmbeddedDB(
        EmbeddedOptions(
            binary_path=str(bin_path),
            version=httpserver.url_for("/tmp_weaviate.tgz"),
            persistence_data_path=tmp_path / "2",
        )
    )
    embedded_db.ensure_weaviate_binary_exists()
    embedded_file_name = list(bin_path.iterdir())
    assert len(embedded_file_name) == 1  # .tgz file was deleted


def test_embedded_ensure_binary_exists_same_as_tar_binary_name(tmp_path):
    bin_path = tmp_path / "notcreated-yet/bin/weaviate"
    assert bin_path.is_file, False
    embedded_db = EmbeddedDB(
        EmbeddedOptions(binary_path=str(bin_path), persistence_data_path=tmp_path)
    )
    embedded_db.ensure_weaviate_binary_exists()
    assert Path(embedded_db.options.binary_path).is_file, True


@pytest.fixture(scope="session")
def embedded_db_binary_path(tmp_path_factory):
    embedded.weaviate_binary_path = (
        tmp_path_factory.mktemp("embedded-test") / "weaviate-embedded-binary"
    )


@pytest.mark.parametrize("options", [EmbeddedOptions(), EmbeddedOptions(port=30666)])
def test_embedded_end_to_end(options, tmp_path):
    options.binary_path = tmp_path
    options.persistence_data_path = tmp_path
    embedded_db = EmbeddedDB(options=options)
    assert embedded_db.is_listening() is False
    with pytest.raises(WeaviateStartUpError):
        with patch("time.sleep") as mocked_sleep:
            embedded_db.wait_till_listening()
            mocked_sleep.assert_has_calls([0.1] * 300)

    embedded_db.ensure_running()
    assert embedded_db.is_listening() is True
    with patch("builtins.print") as mocked_print:
        embedded_db.start()
        mocked_print.assert_called_once_with(
            f"embedded weaviate is already listing on port {options.port}"
        )

    # killing the process should restart it again when ensure running is called
    os.kill(embedded_db.pid, signal.SIGTERM)
    time.sleep(0.2)
    assert embedded_db.is_listening() is False
    embedded_db.ensure_running()
    assert embedded_db.is_listening() is True
    embedded_db.stop()


def test_embedded_multiple_instances(tmp_path):
    embedded_db = EmbeddedDB(
        EmbeddedOptions(
            cluster_hostname="db1",
            port=30664,
            persistence_data_path=(tmp_path / "db1").absolute(),
            binary_path=tmp_path,
        )
    )
    embedded_db2 = EmbeddedDB(
        EmbeddedOptions(
            cluster_hostname="db2",
            port=30665,
            persistence_data_path=(tmp_path / "db2").absolute(),
            binary_path=tmp_path,
        )
    )
    embedded_db.ensure_running()
    assert embedded_db.is_listening() is True
    embedded_db2.ensure_running()
    assert embedded_db2.is_listening() is True


def test_embedded_different_versions(tmp_path):
    client1 = weaviate.Client(
        embedded_options=EmbeddedOptions(
            cluster_hostname="db1",
            port=30664,
            persistence_data_path=(tmp_path / "db1").absolute(),
            binary_path=tmp_path,
            version="https://github.com/weaviate/weaviate/releases/download/v1.18.1/weaviate-v1.18.1-linux-amd64.tar.gz",
        )
    )
    client2 = weaviate.Client(
        embedded_options=EmbeddedOptions(
            cluster_hostname="db2",
            port=30665,
            persistence_data_path=(tmp_path / "db2").absolute(),
            binary_path=tmp_path,
            version="https://github.com/weaviate/weaviate/releases/download/v1.18.0/weaviate-v1.18.0-linux-amd64.tar.gz",
        )
    )
    meta1 = client1.get_meta()
    assert meta1["version"] == "1.18.1"
    meta2 = client2.get_meta()
    assert meta2["version"] == "1.18.0"


def test_custom_env_vars(tmp_path: Path):
    client = weaviate.Client(
        embedded_options=EmbeddedOptions(
            binary_path=tmp_path,
            additional_env_vars={"ENABLE_MODULES": ""},
            persistence_data_path=tmp_path,
        )
    )
    meta = client.get_meta()
    assert len(meta["modules"]) == 0


def test_weaviate_state(tmp_path: Path):
    """Test that weaviate keeps the state between different runs."""
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    port = 36545
    client = weaviate.Client(
        embedded_options=EmbeddedOptions(
            binary_path=tmp_path, port=port, persistence_data_path=tmp_path
        ),
        startup_period=10,
    )
    client.data_object.create({"name": "Name"}, "Person", uuid.uuid4())
    assert sock.connect_ex(("127.0.0.1", port)) == 0  # running
    client._connection.embedded_db.stop()
    del client
    time.sleep(5)  # give weaviate time to shut down

    assert sock.connect_ex(("127.0.0.1", port)) != 0  # not running anymore

    client = weaviate.Client(
        embedded_options=EmbeddedOptions(
            binary_path=tmp_path, port=port, persistence_data_path=tmp_path
        ),
        startup_period=10,
    )
    count = client.query.aggregate("Person").with_meta_count().do()
    assert count["data"]["Aggregate"]["Person"][0]["meta"]["count"] == 1



"""
Cluster class definition.
"""
from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
    EmptyResponseException,
)


class Cluster:
    """
    Cluster class used for cluster information
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Cluster class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection

    def get_nodes_status(self) -> list:
        """
        Get the nodes status.

        Returns
        -------
        list
            List of nodes and their respective status.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        weaviate.EmptyResponseException
            If the response is empty.
        """

        try:
            response = self._connection.get(
                path="/nodes",
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Get nodes status failed due to connection error"
            ) from conn_err

        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Nodes status", response)
        nodes = response.json().get("nodes")
        if nodes is None or nodes == []:
            raise EmptyResponseException("Nodes status response returned empty")
        return nodes



"""
Reference class definition.
"""
import warnings
from typing import Union, Optional

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.error_msgs import (
    REF_DEPRECATION_NEW_V14_CLS_NS_W,
    REF_DEPRECATION_OLD_V14_FROM_CLS_NS_W,
    REF_DEPRECATION_OLD_V14_TO_CLS_NS_W,
)
from weaviate.exceptions import UnexpectedStatusCodeException
from weaviate.util import (
    get_valid_uuid,
    _capitalize_first_letter,
)
from weaviate.data.replication import ConsistencyLevel


class Reference:
    """
    Reference class used to manipulate references within objects.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Reference class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    def delete(
        self,
        from_uuid: str,
        from_property_name: str,
        to_uuid: str,
        from_class_name: Optional[str] = None,
        to_class_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
    ) -> None:
        """
        Remove a reference to another object. Equal to removing one direction of an edge from the
        graph.

        Parameters
        ----------
        from_uuid : str
            The ID of the object that references another object.
        from_property_name : str
            The property from which the reference should be deleted.
        to_uuid : str
            The UUID of the referenced object.
        from_class_name : Optional[str], optional
            The class name of the object for which to delete the reference (with UUID `from_uuid`),
            it is included in Weaviate 1.14.0, where all objects are namespaced by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        to_class_name : Optional[str], optional
            The referenced object class name to which to delete the reference (with UUID `to_uuid`),
            it is included in Weaviate 1.14.0, where all objects are namespaced by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge

        Examples
        --------
        Assume we have two classes, Author and Book.

        >>> # Create the objects first
        >>> client.data_object.create(
        ...     data_object={'name': 'Ray Bradbury'},
        ...     class_name='Author',
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab'
        ... )
        >>> client.data_object.create(
        ...     data_object={'title': 'The Martian Chronicles'},
        ...     class_name='Book',
        ...     uuid='a9c1b714-4f8a-4b01-a930-38b046d69d2d'
        ... )
        >>> # Add the cross references
        >>> ## Author -> Book
        >>> client.data_object.reference.add(
        ...     from_uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     from_property_name='wroteBooks',
        ...     to_uuid='a9c1b714-4f8a-4b01-a930-38b046d69d2d',
        ...     from_class_name='Author', # ONLY with Weaviate >= 1.14.0
        ...     to_class_name='Book', # ONLY with Weaviate >= 1.14.0
        ... )
        >>> client.data_object.get(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617177700595,
            "id": "e067f671-1202-42c6-848b-ff4d1eb804ab",
            "lastUpdateTimeUnix": 1617177700595,
            "properties": {
                "name": "Ray Bradbury",
                "wroteBooks": [
                {
                    "beacon": "weaviate://localhost/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d",
                    "href": "/v1/objects/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d"
                }
                ]
            },
            "vectorWeights": null
        }
        >>> # delete the reference
        >>> client.data_object.reference.delete(
        ...     from_uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     from_property_name='wroteBooks',
        ...     to_uuid='a9c1b714-4f8a-4b01-a930-38b046d69d2d',
        ...     from_class_name='Author', # ONLY with Weaviate >= 1.14.0
        ...     to_class_name='Book', # ONLY with Weaviate >= 1.14.0
        ... )
        >>> >>> client.data_object.get(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617177700595,
            "id": "e067f671-1202-42c6-848b-ff4d1eb804ab",
            "lastUpdateTimeUnix": 1617177864970,
            "properties": {
                "name": "Ray Bradbury",
                "wroteBooks": []
            },
            "vectorWeights": null
        }

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If parameter has the wrong type.
        ValueError
            If uuid is not properly formed.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"
        params = None
        if consistency_level is not None:
            params = {"consistency_level": ConsistencyLevel(consistency_level).value}

        if (from_class_name is None or to_class_name is None) and is_server_version_14:
            warnings.warn(
                message=REF_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if from_class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_FROM_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            _validate_string_arguments(
                argument=from_class_name,
                argument_name="from_class_name",
            )
        if to_class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_TO_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            _validate_string_arguments(
                argument=to_class_name,
                argument_name="to_class_name",
            )

        # Validate and create Beacon
        from_uuid = get_valid_uuid(from_uuid)
        to_uuid = get_valid_uuid(to_uuid)
        _validate_string_arguments(
            argument=from_property_name,
            argument_name="from_property_name",
        )

        if to_class_name and is_server_version_14:
            beacon = _get_beacon(
                to_uuid=to_uuid,
                class_name=_capitalize_first_letter(to_class_name),
            )
        else:
            beacon = _get_beacon(
                to_uuid=to_uuid,
            )

        if from_class_name and is_server_version_14:
            _class_name = _capitalize_first_letter(from_class_name)
            path = f"/objects/{_class_name}/{from_uuid}/references/{from_property_name}"
        else:
            path = f"/objects/{from_uuid}/references/{from_property_name}"

        try:
            response = self._connection.delete(path=path, weaviate_object=beacon, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Reference was not deleted.") from conn_err
        if response.status_code == 204:
            return
        raise UnexpectedStatusCodeException("Delete property reference to object", response)

    def update(
        self,
        from_uuid: str,
        from_property_name: str,
        to_uuids: Union[list, str],
        from_class_name: Optional[str] = None,
        to_class_names: Union[list, str, None] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
    ) -> None:
        """
        Allows to update all references in that property with a new set of references.
        All old references will be deleted.

        Parameters
        ----------
        from_uuid : str
            The object that should have the reference as part of its properties.
            Should be in the form of an UUID or in form of an URL.
            E.g.
            'http://localhost:8080/v1/objects/Book/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'
        from_property_name : str
            The name of the property within the object.
        to_uuids : list or str
            The UUIDs of the objects that should be referenced.
            Should be a list of str in the form of an UUID or str in form of an URL.
            E.g.
            ['http://localhost:8080/v1/objects/Book/fc7eb129-f138-457f-b727-1b29db191a67', ...]
            or
            ['fc7eb129-f138-457f-b727-1b29db191a67', ...]
            If `str` it is converted internally into a list of str.
        from_class_name : Optional[str], optional
            The class name of the object for which to delete the reference (with UUID `from_uuid`),
            it is included in Weaviate 1.14.0, where all objects are namespaced by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        to_class_names : Union[list, str, None], optional
            The referenced objects class name to which to delete the reference (with UUID
            `to_uuid`), it is included in Weaviate 1.14.0, where all objects are namespaced by
            class name. It can be a single class name (assumes all `to_uuids` are of the same
            class) or a list of class names where for each UUID in `to_uuids` we have a class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge

        Examples
        --------
        You have data object 1 with reference property `wroteBooks` and currently has one reference
        to data object 7. Now you say, I want to update the references of data object 1.wroteBooks
        to this list 3,4,9. After the update, the data object 1.wroteBooks is now 3,4,9, but no
        longer contains 7.

        >>> client.data_object.get(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab'
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617177700595,
            "id": "e067f671-1202-42c6-848b-ff4d1eb804ab",
            "lastUpdateTimeUnix": 1617177700595,
            "properties": {
                "name": "Ray Bradbury",
                "wroteBooks": [
                {
                    "beacon": "weaviate://localhost/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d",
                    "href": "/v1/objects/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d"
                }
                ]
            },
            "vectorWeights": null
        }
        Currently there is only one `Book` reference.
        Update all the references of the Author for property name `wroteBooks`.
        >>> client.data_object.reference.update(
        ...     from_uuid = 'e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     from_property_name = 'wroteBooks',
        ...     to_uuids = [
        ...         '8429f68f-860a-49ea-a50b-1f8789515882',
        ...         '3e2e6795-298b-47e9-a2cb-3d8a77a24d8a'
        ...     ],
        ...     from_class_name='Author', # ONLY with Weaviate >= 1.14.0
        ...     to_class_name='Book', # ONLY with Weaviate >= 1.14.0
        ... )
        >>> client.data_object.get(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab'
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617181292677,
            "id": "e067f671-1202-42c6-848b-ff4d1eb804ab",
            "lastUpdateTimeUnix": 1617181409405,
            "properties": {
                "name": "Ray Bradbury",
                "wroteBooks": [
                {
                    "beacon": "weaviate://localhost/Book/8429f68f-860a-49ea-a50b-1f8789515882",
                    "href": "/v1/objects/Book/8429f68f-860a-49ea-a50b-1f8789515882"
                },
                {
                    "beacon": "weaviate://localhost/Book/3e2e6795-298b-47e9-a2cb-3d8a77a24d8a",
                    "href": "/v1/objects/Book/3e2e6795-298b-47e9-a2cb-3d8a77a24d8a"
                }
                ]
            },
            "vectorWeights": null
        }
        All the previous references were removed and now we have only those specified in the
        `update` method.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If the parameters are of the wrong type.
        ValueError
            If the parameters are of the wrong value.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"
        params = None
        if consistency_level is not None:
            params = {"consistency_level": ConsistencyLevel(consistency_level).value}

        if (from_class_name is None or to_class_names is None) and is_server_version_14:
            warnings.warn(
                message=REF_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if from_class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_FROM_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            _validate_string_arguments(
                argument=from_class_name,
                argument_name="from_class_name",
            )
        if to_class_names is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_TO_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )

            if not isinstance(to_class_names, list):
                _validate_string_arguments(
                    argument=to_class_names,
                    argument_name="to_class_names",
                )
            else:
                for to_class_name in to_class_names:
                    if not isinstance(to_class_name, str):
                        raise TypeError(
                            "'to_class_names' must be of type str of List[str]. "
                            f"Found element of type: {type(to_class_name)}"
                        )
                if len(to_class_names) == 0:
                    to_class_names = None

        if not isinstance(to_uuids, list):
            to_uuids = [to_uuids]
        if isinstance(to_class_names, str):
            to_class_names = [to_class_names] * len(to_uuids)
        if to_class_names is not None and len(to_uuids) != len(to_class_names):
            raise ValueError(
                "'to_class_names' and 'to_uuids' have different lengths, they must match."
            )

        # Validate and create Beacon
        from_uuid = get_valid_uuid(from_uuid)
        _validate_string_arguments(
            argument=from_property_name,
            argument_name="from_property_name",
        )
        beacons = []

        if to_class_names and is_server_version_14:
            for to_uuid, to_class_name in zip(to_uuids, to_class_names):
                to_uuid = get_valid_uuid(to_uuid)
                beacon = _get_beacon(
                    to_uuid=to_uuid,
                    class_name=_capitalize_first_letter(to_class_name),
                )
                beacons.append(beacon)
        else:
            for to_uuid in to_uuids:
                to_uuid = get_valid_uuid(to_uuid)
                beacon = _get_beacon(
                    to_uuid=to_uuid,
                )
                beacons.append(beacon)

        if from_class_name and is_server_version_14:
            _class_name = _capitalize_first_letter(from_class_name)
            path = f"/objects/{_class_name}/{from_uuid}/references/{from_property_name}"
        else:
            path = f"/objects/{from_uuid}/references/{from_property_name}"

        try:
            response = self._connection.put(
                path=path,
                weaviate_object=beacons,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Reference was not updated.") from conn_err
        if response.status_code == 200:
            return
        raise UnexpectedStatusCodeException("Update property reference to object", response)

    def add(
        self,
        from_uuid: str,
        from_property_name: str,
        to_uuid: str,
        from_class_name: Optional[str] = None,
        to_class_name: Optional[str] = None,
        consistency_level: Optional[ConsistencyLevel] = None,
    ) -> None:
        """
        Allows to link an object to an object uni-directionally.

        Parameters
        ----------
        from_uuid : str
            The ID of the object that should have the reference as part
            of its properties. Should be a plane UUID or an URL.
            E.g.
            'http://localhost:8080/v1/objects/Book/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'
        from_property_name : str
            The name of the property within the object.
        to_uuid : str
            The UUID of the object that should be referenced.
            Should be a plane UUID or an URL.
            E.g.
            'http://localhost:8080/v1/objects/Book/fc7eb129-f138-457f-b727-1b29db191a67'
            or
            'fc7eb129-f138-457f-b727-1b29db191a67'
        from_class_name : Optional[str], optional
            The class name of the object for which to delete the reference (with UUID `from_uuid`),
            it is included in Weaviate 1.14.0, where all objects are namespaced by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        to_class_name : Optional[str], optional
            The referenced object class name to which to delete the reference (with UUID `to_uuid`),
            it is included in Weaviate 1.14.0, where all objects are namespaced by class name.
            STRONGLY recommended to set it with Weaviate >= 1.14.0. It will be required in future
            versions of Weaviate Server and Clients. Use None value ONLY for Weaviate < v1.14.0,
            by default None
        consistency_level : Optional[ConsistencyLevel], optional
            Can be one of 'ALL', 'ONE', or 'QUORUM'. Determines how many replicas must acknowledge

        Examples
        --------
        Assume we have two classes, Author and Book.

        >>> # Create the objects first
        >>> client.data_object.create(
        ...     data_object={'name': 'Ray Bradbury'},
        ...     class_name='Author',
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab'
        ... )
        >>> client.data_object.create(
        ...     data_object={'title': 'The Martian Chronicles'},
        ...     class_name='Book',
        ...     uuid='a9c1b714-4f8a-4b01-a930-38b046d69d2d'
        ... )
        >>> # Add the cross references
        >>> ## Author -> Book
        >>> client.data_object.reference.add(
        ...     from_uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     from_property_name='wroteBooks',
        ...     to_uuid='a9c1b714-4f8a-4b01-a930-38b046d69d2d',
        ...     from_class_name='Author', # ONLY with Weaviate >= 1.14.0
        ...     to_class_name='Book', # ONLY with Weaviate >= 1.14.0
        ... )
        >>> client.data_object.get(
        ...     uuid='e067f671-1202-42c6-848b-ff4d1eb804ab',
        ...     class_name='Author', # ONLY with Weaviate >= 1.14.0
        ... )
        {
            "additional": {},
            "class": "Author",
            "creationTimeUnix": 1617177700595,
            "id": "e067f671-1202-42c6-848b-ff4d1eb804ab",
            "lastUpdateTimeUnix": 1617177700595,
            "properties": {
                "name": "Ray Bradbury",
                "wroteBooks": [
                {
                    "beacon": "weaviate://localhost/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d",
                    "href": "/v1/objects/Book/a9c1b714-4f8a-4b01-a930-38b046d69d2d"
                }
                ]
            },
            "vectorWeights": null
        }

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        TypeError
            If the parameters are of the wrong type.
        ValueError
            If the parameters are of the wrong value.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"
        params = None
        if consistency_level is not None:
            params = {"consistency_level": ConsistencyLevel(consistency_level).value}

        if (from_class_name is None or to_class_name is None) and is_server_version_14:
            warnings.warn(
                message=REF_DEPRECATION_NEW_V14_CLS_NS_W,
                category=DeprecationWarning,
                stacklevel=1,
            )
        if from_class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_FROM_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            _validate_string_arguments(
                argument=from_class_name,
                argument_name="from_class_name",
            )
        if to_class_name is not None:
            if not is_server_version_14:
                warnings.warn(
                    message=REF_DEPRECATION_OLD_V14_TO_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )
            _validate_string_arguments(
                argument=to_class_name,
                argument_name="to_class_name",
            )

        # Validate and create Beacon
        from_uuid = get_valid_uuid(from_uuid)
        to_uuid = get_valid_uuid(to_uuid)
        _validate_string_arguments(
            argument=from_property_name,
            argument_name="from_property_name",
        )

        if to_class_name and is_server_version_14:
            beacon = _get_beacon(
                to_uuid=to_uuid,
                class_name=_capitalize_first_letter(to_class_name),
            )
        else:
            beacon = _get_beacon(
                to_uuid=to_uuid,
            )

        if from_class_name and is_server_version_14:
            _class_name = _capitalize_first_letter(from_class_name)
            path = f"/objects/{_class_name}/{from_uuid}/references/{from_property_name}"
        else:
            path = f"/objects/{from_uuid}/references/{from_property_name}"

        try:
            response = self._connection.post(
                path=path,
                weaviate_object=beacon,
                params=params,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Reference was not added.") from conn_err
        if response.status_code == 200:
            return
        raise UnexpectedStatusCodeException("Add property reference to object", response)


def _get_beacon(to_uuid: str, class_name: Optional[str] = None) -> dict:
    """
    Get a weaviate-style beacon.

    Parameters
    ----------
    to_uuid : str
        The UUID to create beacon for.
    class_name : Optional[str], optional
        The class name of the `to_uuid` object. Used with Weaviate >= 1.14.0.
        For Weaviate < 1.14.0 use None value.

    Returns
    -------
    dict
        Weaviate-style beacon as a dict.
    """

    if class_name is None:
        return {"beacon": f"weaviate://localhost/{to_uuid}"}
    return {"beacon": f"weaviate://localhost/{class_name}/{to_uuid}"}


def _validate_string_arguments(argument: str, argument_name: str) -> None:
    """
    Validate string arguments.

    Parameters
    ----------
    argument : str
        Argument value to be validated.
    argument_name : str
        Argument name to be included in error message.

    Raises
    ------
    TypeError
        If 'argument' is not of type str.
    """

    if not isinstance(argument, str):
        raise TypeError(f"'{argument_name}' must be of type 'str'. Given type: {type(argument)}")



"""
Weaviate Python Client Library used to interact with a Weaviate instance.

The interaction with Weaviate instance should be through a `Client` object. A `Client` instance
has instance attributes to all the object needed to create objects/schema, do classification,
upload batches, query data, ... Creating separate `Schema`, `DataObject`, `Batch`,
`Classification`, `Query`, `Connect`, `Reference` is **STRONGLY DISCOURAGED**. The `Client` class
creates the needed instances and connects all of them to the same Weaviate instance for you.

Examples
--------
Creating and exploring a Weaviate instance running on `localhost`, on port `8080`, with Authentication disabled.

>>> import weaviate
>>> client = weaviate.Client('http://localhost:8080')
>>> print_type = lambda obj: print(type(obj))
>>> print_type(client.batch)
<class 'weaviate.batch.crud_batch.Batch'>
>>> print_type(client.schema)
<class 'weaviate.schema.crud_schema.Schema'>
>>> print_type(client.classification)
<class 'weaviate.classification.classify.Classification'>
>>> print_type(client.data_object)
<class 'weaviate.data.crud_data.DataObject'>
>>> print_type(client.query)
<class 'weaviate.gql.query.Query'>

Attributes
----------
__version__ : str
    Current `weaviate-python` library version installed.
"""


__all__ = [
    "Client",
    "AuthClientCredentials",
    "AuthClientPassword",
    "AuthBearerToken",
    "AuthApiKey",
    "UnexpectedStatusCodeException",
    "ObjectAlreadyExistsException",
    "AuthenticationFailedException",
    "SchemaValidationException",
    "ConsistencyLevel",
    "WeaviateErrorRetryConf",
]

import sys

from .auth import AuthClientCredentials, AuthClientPassword, AuthBearerToken, AuthApiKey
from .batch.crud_batch import WeaviateErrorRetryConf
from .client import Client
from .data.replication import ConsistencyLevel
from .exceptions import (
    UnexpectedStatusCodeException,
    ObjectAlreadyExistsException,
    AuthenticationFailedException,
    SchemaValidationException,
)
from .version import __version__  # noqa

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("default")



"""
Helper functions!
"""
import base64
import json
import os
import uuid as uuid_lib
from enum import Enum, EnumMeta
from io import BufferedReader
from numbers import Real
from typing import Union, Sequence, Any, Optional, List, Dict

import requests
import validators

from weaviate.exceptions import SchemaValidationException


# MetaEnum and BaseEnum are required to support `in` statements:
#    'ALL' in ConsistencyLevel == True
#    12345 in ConsistencyLevel == False
class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            # when item is type ConsistencyLevel
            return item.name in cls.__members__.keys()
        except AttributeError:
            # when item is type str
            return item in cls.__members__.keys()


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


def image_encoder_b64(image_or_image_path: Union[str, BufferedReader]) -> str:
    """
    Encode a image in a Weaviate understandable format from a binary read file or by providing
    the image path.

    Parameters
    ----------
    image_or_image_path : str, io.BufferedReader
        The binary read file or the path to the file.

    Returns
    -------
    str
        Encoded image.

    Raises
    ------
    ValueError
        If the argument is str and does not point to an existing file.
    TypeError
        If the argument is of a wrong data type.
    """

    if isinstance(image_or_image_path, str):
        if not os.path.isfile(image_or_image_path):
            raise ValueError("No file found at location " + image_or_image_path)
        with open(image_or_image_path, "br") as file:
            content = file.read()

    elif isinstance(image_or_image_path, BufferedReader):
        content = image_or_image_path.read()
    else:
        raise TypeError(
            '"image_or_image_path" should be a image path or a binary read file'
            " (io.BufferedReader)"
        )
    return base64.b64encode(content).decode("utf-8")


def image_decoder_b64(encoded_image: str) -> bytes:
    """
    Decode image from a Weaviate format image.

    Parameters
    ----------
    encoded_image : str
        The encoded image.

    Returns
    -------
    bytes
        Decoded image as a binary string.
    """

    return base64.b64decode(encoded_image.encode("utf-8"))


def generate_local_beacon(
    to_uuid: Union[str, uuid_lib.UUID],
    class_name: Optional[str] = None,
) -> dict:
    """
    Generates a beacon with the given uuid and class name (only for Weaviate >= 1.14.0).

    Parameters
    ----------
    to_uuid : str or uuid.UUID
        The UUID for which to create a local beacon.
    class_name : Optional[str], optional
        The class name of the `to_uuid` object. Used with Weaviate >= 1.14.0.
        For Weaviate < 1.14.0 use None value.

    Returns
    -------
    dict
        The local beacon.

    Raises
    ------
    TypeError
        If 'to_uuid' is not of type str.
    ValueError
        If the 'to_uuid' is not valid.
    """

    if isinstance(to_uuid, str):
        try:
            uuid = str(uuid_lib.UUID(to_uuid))
        except ValueError:
            raise ValueError("Uuid does not have the proper form") from None
    elif isinstance(to_uuid, uuid_lib.UUID):
        uuid = str(to_uuid)
    else:
        raise TypeError("Expected to_object_uuid of type str or uuid.UUID")

    if class_name is None:
        return {"beacon": f"weaviate://localhost/{uuid}"}
    return {"beacon": f"weaviate://localhost/{class_name}/{uuid}"}


def _get_dict_from_object(object_: Union[str, dict]) -> dict:
    """
    Takes an object that should describe a dict
    e.g. a schema or an object and tries to retrieve the dict.

    Parameters
    ----------
    object_ : str or dict
        The object from which to retrieve the dict.
        Can be a python dict, or the path to a json file or a url of a json file.

    Returns
    -------
    dict
        The object as a dict.

    Raises
    ------
    TypeError
        If 'object_' is neither a string nor a dict.
    ValueError
        If no dict can be retrieved from object.
    """

    # check if objects files is url
    if object_ is None:
        raise TypeError("argument is None")

    if isinstance(object_, dict):
        # Object is already a dict
        return object_
    if isinstance(object_, str):
        if validators.url(object_):
            # Object is URL
            response = requests.get(object_)
            if response.status_code == 200:
                return response.json()
            raise ValueError("Could not download file " + object_)

        if not os.path.isfile(object_):
            # Object is neither file nor URL
            raise ValueError("No file found at location " + object_)
        # Object is file
        with open(object_, "r") as file:
            return json.load(file)
    raise TypeError(
        "Argument is not of the supported types. Supported types are "
        "url or file path as string or schema as dict."
    )


def is_weaviate_object_url(url: str) -> bool:
    """
    Checks if the input follows a normal Weaviate 'beacon' like this:
    'weaviate://localhost/ClassName/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Parameters
    ----------
    url : str
        The URL to be validated.

    Returns
    -------
    bool
        True if the 'url' is a Weaviate object URL.
        False otherwise.
    """

    if not isinstance(url, str):
        return False
    if not url.startswith("weaviate://"):
        return False
    url = url[11:]
    split = url.split("/")
    if len(split) not in (2, 3):
        return False
    if split[0] != "localhost":
        if not validators.domain(split[0]):
            return False
    try:
        uuid_lib.UUID(split[-1])
    except ValueError:
        return False
    return True


def is_object_url(url: str) -> bool:
    """
    Validates an url like 'http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446'
    or '/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446' references a object. It only validates
    the path format and UUID, not the host or the protocol.

    Parameters
    ----------
    url : str
        The URL to be validated.

    Returns
    -------
    bool
        True if the 'url' is a valid path to an object.
        False otherwise.
    """

    v1_split = url.split("/v1/")

    if len(v1_split) != 2:
        return False

    split = v1_split[1].split("/")

    if len(split) not in (2, 3):
        return False

    try:
        uuid_lib.UUID(split[-1])
    except ValueError:
        return False
    if not split[0] == "objects":
        return False
    return True


def get_valid_uuid(uuid: Union[str, uuid_lib.UUID]) -> str:
    """
    Validate and extract the UUID.

    Parameters
    ----------
    uuid : str or uuid.UUID
        The UUID to be validated and extracted.
        Should be in the form of an UUID or in form of an URL (weaviate 'beacon' or 'href').
        E.g.
        'http://localhost:8080/v1/objects/fc7eb129-f138-457f-b727-1b29db191a67'
        or
        'weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'
        or
        'fc7eb129-f138-457f-b727-1b29db191a67'

    Returns
    -------
    str
        The extracted UUID.

    Raises
    ------
    TypeError
        If 'uuid' is not of type str.
    ValueError
        If 'uuid' is not valid or cannot be extracted.
    """

    if isinstance(uuid, uuid_lib.UUID):
        return str(uuid)

    if not isinstance(uuid, str):
        raise TypeError("'uuid' must be of type str or uuid.UUID, but was: " + str(type(uuid)))

    _is_weaviate_url = is_weaviate_object_url(uuid)
    _is_object_url = is_object_url(uuid)
    _uuid = uuid
    if _is_weaviate_url or _is_object_url:
        _uuid = uuid.split("/")[-1]
    try:
        _uuid = str(uuid_lib.UUID(_uuid))
    except ValueError:
        raise ValueError("Not valid 'uuid' or 'uuid' can not be extracted from value") from None
    return _uuid


def get_vector(vector: Sequence) -> list:
    """
    Get weaviate compatible format of the embedding vector.

    Parameters
    ----------
    vector: Sequence
        The embedding of an object. Used only for class objects that do not have a vectorization
        module. Supported types are `list`, `numpy.ndarray`, `torch.Tensor` and `tf.Tensor`.

    Returns
    -------
    list
        The embedding as a list.

    Raises
    ------
    TypeError
        If 'vector' is not of a supported type.
    """

    if isinstance(vector, list):
        # if vector is already a list
        return vector
    try:
        # if vector is numpy.ndarray or torch.Tensor
        return vector.squeeze().tolist()
    except AttributeError:
        try:
            # if vector is tf.Tensor
            return vector.numpy().squeeze().tolist()
        except AttributeError:
            raise TypeError(
                "The type of the 'vector' argument is not supported!\n"
                "Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` and `tf.Tensor`"
            ) from None


def get_domain_from_weaviate_url(url: str) -> str:
    """
    Get the domain from a weaviate URL.

    Parameters
    ----------
    url : str
        The weaviate URL.
        Of this form: 'weaviate://localhost/objects/28f3f61b-b524-45e0-9bbe-2c1550bf73d2'

    Returns
    -------
    str
        The domain.
    """

    return url[11:].split("/")[0]


def _is_sub_schema(sub_schema: dict, schema: dict) -> bool:
    """
    Check for a subset in a schema.

    Parameters
    ----------
    sub_schema : dict
        The smaller schema that should be contained in the 'schema'.
    schema : dict
        The schema for which to check if 'sub_schema' is a part of. Must have the 'classes' key.

    Returns
    -------
    bool
        True is 'sub_schema' is a subset of the 'schema'.
        False otherwise.
    """

    schema_classes = schema.get("classes", [])
    if "classes" in sub_schema:
        sub_schema_classes = sub_schema["classes"]
    else:
        sub_schema_classes = [sub_schema]
    return _compare_class_sets(sub_schema_classes, schema_classes)


def _compare_class_sets(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of classes.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    set_ : list
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_class in sub_set:
        found = False
        for set_class in set_:
            if "class" not in sub_set_class:
                raise SchemaValidationException(
                    "The sub schema class/es MUST have a 'class' keyword each!"
                )
            if _capitalize_first_letter(sub_set_class["class"]) == _capitalize_first_letter(
                set_class["class"]
            ):
                if _compare_properties(sub_set_class["properties"], set_class["properties"]):
                    found = True
                    break
        if not found:
            return False
    return True


def _compare_properties(sub_set: list, set_: list) -> bool:
    """
    Check for a subset in a set of properties.

    Parameters
    ----------
    sub_set : list
        The smaller set that should be contained in the 'set'.
    set_ : list
        The set for which to check if 'sub_set' is a part of.

    Returns
    -------
    bool
        True is 'sub_set' is a subset of the 'set'.
        False otherwise.
    """

    for sub_set_property in sub_set:
        found = False
        for set_property in set_:
            if sub_set_property["name"] == set_property["name"]:
                found = True
                break
        if not found:
            return False
    return True


def generate_uuid5(identifier: Any, namespace: Any = "") -> str:
    """
    Generate an UUIDv5, may be used to consistently generate the same UUID for a specific
    identifier and namespace.

    Parameters
    ----------
    identifier : Any
        The identifier/object that should be used as basis for the UUID.
    namespace : Any, optional
        Allows to namespace the identifier, by default ""

    Returns
    -------
    str
        The UUID as a string.
    """

    return str(uuid_lib.uuid5(uuid_lib.NAMESPACE_DNS, str(namespace) + str(identifier)))


def _capitalize_first_letter(string: str) -> str:
    """
    Capitalize only the first letter of the `string`.

    Parameters
    ----------
    string : str
        The string to be capitalized.

    Returns
    -------
    str
        The capitalized string.
    """

    if len(string) == 1:
        return string.capitalize()
    return string[0].capitalize() + string[1:]


def check_batch_result(
    results: Optional[List[Dict[str, Any]]],
) -> None:
    """
    Check batch results for errors.

    Parameters
    ----------
    results : dict
        The Weaviate batch creation return value.
    """

    if results is None:
        return
    for result in results:
        if "result" in result and "errors" in result["result"]:
            if "error" in result["result"]["errors"]:
                print(result["result"]["errors"])


def _check_positive_num(
    value: Real, arg_name: str, data_type: type, include_zero: bool = False
) -> None:
    """
    Check if the `value` of the `arg_name` is a positive number.

    Parameters
    ----------
    value : Union[int, float]
        The value to check.
    arg_name : str
        The name of the variable from the original function call. Used for error message.
    data_type : type
        The data type to check for.
    include_zero : bool
        Wether zero counts as positive or not. By default False.

    Raises
    ------
    TypeError
        If the `value` is not of type `data_type`.
    ValueError
        If the `value` has a non positive value.
    """

    if not isinstance(value, data_type) or isinstance(value, bool):
        raise TypeError(f"'{arg_name}' must be of type {data_type}.")
    if include_zero:
        if value < 0:
            raise ValueError(f"'{arg_name}' must be positive, i.e. greater or equal to zero (>=0).")
    else:
        if value <= 0:
            raise ValueError(f"'{arg_name}' must be positive, i.e. greater that zero (>0).")


def is_weaviate_domain(url: str) -> bool:
    return (
        "weaviate.io" in url.lower()
        or "semi.technology" in url.lower()
        or "weaviate.cloud" in url.lower()
    )



"""
Schema validation module.
"""
from typing import Any

from weaviate.exceptions import SchemaValidationException

CLASS_KEYS = {
    "class",
    "vectorIndexType",
    "vectorIndexConfig",
    "moduleConfig",
    "description",
    "vectorizer",
    "properties",
    "invertedIndexConfig",
    "shardingConfig",
    "replicationConfig",
}

PROPERTY_KEYS = {"dataType", "name", "moduleConfig", "description", "indexInverted", "tokenization"}


def validate_schema(schema: dict) -> None:
    """
    Validate schema.

    Parameters
    ----------
    schema : dict
        Schema to be validated.

    Raises
    ------
    weaviate.SchemaValidationException
        If the schema could not be validated against the standard format.
    """

    # check if schema has required "classes" as keys
    if "classes" not in schema:
        raise SchemaValidationException(
            'Each schema has to have "classes" '
            "in the first level of the JSON format file/parameter/object"
        )
    # check if "classes" is of type list
    _check_key_type("classes", schema["classes"], list)
    # check if each class in the "classes" is a valid class
    for weaviate_class in schema["classes"]:
        _check_key_type("class", weaviate_class, dict)
        check_class(weaviate_class)


def check_class(class_definition: dict) -> None:
    """
    Validate a class against the standard class format.

    Parameters
    ----------
    class_definition : dict
        The definition of the class to be validated.

    Raises
    ------
    weaviate.SchemaValidationException
        If the class could not be validated against the standard class format.
    """

    # check mandatory keys
    if "class" not in class_definition:
        raise SchemaValidationException('"class" key is missing in class definition.')

    # check optional keys
    for key in class_definition.keys():
        # Check if key is known
        if key not in CLASS_KEYS:
            raise SchemaValidationException(f'"{key}" is not a known class definition key.')
        # check if key is right type
        if key in ["class", "vectorIndexType", "description", "vectorizer"]:
            _check_key_type(key, class_definition[key], str)
        if key in [
            "vectorIndexConfig",
            "moduleConfig",
            "invertedIndexConfig",
            "shardingConfig",
            "replicationConfig",
        ]:
            _check_key_type(key, class_definition[key], dict)
        if key in ["properties"]:
            _check_key_type(key, class_definition[key], list)

    if "properties" in class_definition:
        for class_property in class_definition["properties"]:
            check_property(class_property)


def check_property(class_property: dict) -> None:
    """
    Validate a class property against the standard class property.

    Parameters
    ----------
    class_property : dict
        The class property to be validated.

    Raises
    ------
    weaviate.SchemaValidationException
        If the class property could not be validated against the standard class property format.
    """

    # mandatory fields
    if "dataType" not in class_property:
        raise SchemaValidationException('Property does not contain "dataType"')
    if "name" not in class_property:
        raise SchemaValidationException('Property does not contain "name"')

    for key in class_property:
        # check for misspelled and/or non-existent properties
        if key not in PROPERTY_KEYS:
            raise SchemaValidationException(f'Property "{key}" is not known.')

        # Test types
        if key in ["dataType"]:
            _check_key_type(key, class_property[key], list)
        if key in ["name", "description", "tokenization"]:
            _check_key_type(key, class_property[key], str)
        if key in ["indexInverted"]:
            _check_key_type(key, class_property[key], bool)
        if key in ["moduleConfig"]:
            _check_key_type(key, class_property[key], dict)

    # Test dataType types
    for data_type in class_property["dataType"]:
        _check_key_type("dataType object", data_type, str)


def _check_key_type(key: str, value: Any, expected_type: Any) -> None:
    """
    Check if value is of an expected type.

    Parameters
    ----------
    key : str
        The key for which to check data type.
    value : Any
        The value of the 'key' for which to check data type.
    expected_type : Any
        The expected data type of the 'value'.

    Raises
    ------
    weaviate.SchemaValidationException
        If the 'value' is of wrong data type.
    """

    if not isinstance(value, expected_type):
        raise SchemaValidationException(
            f'"{key}" is type {type(value)} ' f"but should be {expected_type}."
        )



"""
GraphQL query module.
"""
from typing import List, Union

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException
from .aggregate import AggregateBuilder
from .get import GetBuilder
from .multi_get import MultiGetBuilder


class Query:
    """
    Query class used to make `get` and/or `aggregate` GraphQL queries.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a Classification class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.
        """

        self._connection = connection

    def get(
        self,
        class_name: str,
        properties: Union[List[str], str, None] = None,
    ) -> GetBuilder:
        """
        Instantiate a GetBuilder for GraphQL `get` requests.

        Parameters
        ----------
        class_name : str
            Class name of the objects to interact with.
        properties : list of str, str or None
            Properties of the objects to get, by default None

        Returns
        -------
        GetBuilder
            A GetBuilder to make GraphQL `get` requests from weaviate.
        """

        return GetBuilder(class_name, properties, self._connection)

    def multi_get(
        self,
        get_builder: List[GetBuilder],
    ) -> MultiGetBuilder:
        """
        Instantiate a MultiGetBuilder for GraphQL `multi_get` requests.
        Bundels multiple get requests into one.

        Parameters
        ----------
        get_builder : list of GetBuilder
            List of GetBuilder objects for a single request each.

        Returns
        -------
        MultiGetBuilder
            A MultiGetBuilder to make GraphQL `get` multiple requests from weaviate.
        """

        return MultiGetBuilder(get_builder, self._connection)

    def aggregate(self, class_name: str) -> AggregateBuilder:
        """
        Instantiate an AggregateBuilder for GraphQL `aggregate` requests.

        Parameters
        ----------
        class_name : str
            Class name of the objects to be aggregated.

        Returns
        -------
        AggregateBuilder
            An AggregateBuilder to make GraphQL `aggregate` requests from weaviate.
        """

        return AggregateBuilder(class_name, self._connection)

    def raw(self, gql_query: str) -> dict:
        """
        Allows to send simple graph QL string queries.
        Be cautious of injection risks when generating query strings.

        Parameters
        ----------
        gql_query : str
            GraphQL query as a string.

        Returns
        -------
        dict
            Data response of the query.

        Examples
        --------
        >>> query = \"""
        ... {
        ...     Get {
        ...         Article(limit: 2) {
        ...         title
        ...         hasAuthors {
        ...             ... on Author {
        ...                 name
        ...                 }
        ...             }
        ...         }
        ...     }
        ... }
        ... \"""
        >>> client.query.raw(query)
        {
        "data": {
            "Get": {
            "Article": [
                {
                "hasAuthors": [
                    {
                    "name": "Jonathan Wilson"
                    }
                ],
                "title": "Sergio Ag\u00fcero has been far more than a great goalscorer for
                            Manchester City"
                },
                {
                "hasAuthors": [
                    {
                    "name": "Emma Elwick-Bates"
                    }
                ],
                "title": "At Swarovski, Giovanna Engelbert Is Crafting Jewels As Exuberantly
                            Joyful As She Is"
                }
            ]
            }
        },
        "errors": null
        }

        Raises
        ------
        TypeError
            If 'gql_query' is not of type str.
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """

        if not isinstance(gql_query, str):
            raise TypeError("Query is expected to be a string")

        json_query = {"query": gql_query}

        try:
            response = self._connection.post(path="/graphql", weaviate_object=json_query)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Query not executed.") from conn_err
        if response.status_code == 200:
            return response.json()  # Successfully queried
        raise UnexpectedStatusCodeException("GQL query failed", response)



import unittest
import uuid as uuid_lib
from copy import deepcopy
from unittest.mock import patch, Mock

from test.util import check_error_message
from weaviate import SchemaValidationException
from weaviate.util import (
    generate_uuid5,
    image_decoder_b64,
    image_encoder_b64,
    generate_local_beacon,
    is_object_url,
    is_weaviate_object_url,
    get_vector,
    get_valid_uuid,
    get_domain_from_weaviate_url,
    _get_dict_from_object,
    _is_sub_schema,
)

schema_set = {
    "classes": [
        {"class": "Ollie", "properties": [{"name": "height"}]},
        {"class": "Shuvit", "properties": [{"name": "direction"}]},
        {"class": "Board", "properties": [{"name": "brand"}, {"name": "art"}, {"name": "size"}]},
        {"class": "Truck", "properties": [{"name": "name"}, {"name": "height"}]},
    ],
}

schema_set_extended_prop = {
    "classes": [
        {"class": "Ollie", "properties": [{"name": "weight"}]},
        {"class": "Shuvit", "properties": [{"name": "direction"}]},
        {"class": "Board", "properties": [{"name": "brand"}, {"name": "art"}, {"name": "size"}]},
        {"class": "Truck", "properties": [{"name": "name"}, {"name": "height"}]},
    ],
}

schema_sub_set = {
    "classes": [
        {"class": "Ollie", "properties": [{"name": "height"}]},
        {"class": "Board", "properties": [{"name": "brand"}, {"name": "art"}, {"name": "size"}]},
    ],
}

disjoint_set = {
    "classes": [
        {"class": "Manual", "properties": [{"name": "nose"}]},
        {"class": "Bearings", "properties": [{"name": "brand"}]},
    ],
}

partial_set = {
    "classes": [
        {"class": "Board", "properties": [{"name": "brand"}, {"name": "art"}, {"name": "size"}]},
        {"class": "Truck", "properties": [{"name": "name"}, {"name": "height"}]},
        {"class": "Bearings", "properties": [{"name": "brand"}]},
        {"class": "Ollie", "properties": [{"name": "height"}]},
        {"class": "Shuvit", "properties": [{"name": "direction"}]},
        {"class": "Manual", "properties": [{"name": "nose"}]},
    ],
}

schema_company = {
    "classes": [
        {
            "class": "Company",
            "description": "A business that acts in the market",
            "properties": [
                {
                    "name": "name",
                    "description": "The name under which the company is known",
                    "dataType": ["text"],
                },
                {
                    "name": "legalBody",
                    "description": "The legal body under which the company maintains its business",
                    "dataType": ["text"],
                },
                {
                    "name": "hasEmployee",
                    "description": "The employees of the company",
                    "dataType": ["Employee"],
                },
            ],
        },
        {
            "class": "Employee",
            "description": "An employee of the company",
            "properties": [
                {"name": "name", "description": "The name of the employee", "dataType": ["text"]},
                {
                    "name": "job",
                    "description": "the job description of the employee",
                    "dataType": ["text"],
                },
                {
                    "name": "yearsInTheCompany",
                    "description": "The number of years this employee has worked in the company",
                    "dataType": ["int"],
                },
            ],
        },
    ]
}


class TestUtil(unittest.TestCase):
    def test_generate_local_beacon(self):
        """
        Test the `generate_local_beacon` function.
        """

        type_error_message = "Expected to_object_uuid of type str or uuid.UUID"
        value_error_message = "Uuid does not have the proper form"
        # wrong data type
        with self.assertRaises(TypeError) as error:
            generate_local_beacon(None)
        check_error_message(self, error, type_error_message)
        # wrong value
        with self.assertRaises(ValueError) as error:
            generate_local_beacon("Leeroy Jenkins")
        check_error_message(self, error, value_error_message)

        beacon = generate_local_beacon("fcf33178-1b5d-5174-b2e7-04a2129dd35a")
        self.assertTrue("beacon" in beacon)
        self.assertEqual(
            beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35a"
        )

        beacon = generate_local_beacon("fcf33178-1b5d-5174-b2e7-04a2129dd35b")
        self.assertTrue("beacon" in beacon)
        self.assertEqual(
            beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35b"
        )

        beacon = generate_local_beacon("fcf331781b5d5174b2e704a2129dd35b")
        self.assertTrue("beacon" in beacon)
        self.assertEqual(
            beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35b"
        )

        beacon = generate_local_beacon(uuid_lib.UUID("fcf33178-1b5d-5174-b2e7-04a2129dd35b"))
        self.assertTrue("beacon" in beacon)
        self.assertEqual(
            beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35b"
        )

        beacon = generate_local_beacon(uuid_lib.UUID("fcf331781b5d5174b2e704a2129dd35b"))
        self.assertTrue("beacon" in beacon)
        self.assertEqual(
            beacon["beacon"], "weaviate://localhost/fcf33178-1b5d-5174-b2e7-04a2129dd35b"
        )

    def test__get_dict_from_object(self):
        """
        Test the `_get_dict_from_object` function.
        """

        none_error_message = "argument is None"
        file_error_message = "No file found at location "
        url_error_message = "Could not download file "
        type_error_message = (
            "Argument is not of the supported types. Supported types are "
            "url or file path as string or schema as dict."
        )
        # test wrong type None
        with self.assertRaises(TypeError) as error:
            _get_dict_from_object(None)
        check_error_message(self, error, none_error_message)
        # wrong data type
        with self.assertRaises(TypeError) as error:
            _get_dict_from_object([{"key": 1234}])
        check_error_message(self, error, type_error_message)
        # wrong path
        with self.assertRaises(ValueError) as error:
            _get_dict_from_object("not_a_path_or_url.txt")
        check_error_message(self, error, file_error_message + "not_a_path_or_url.txt")
        # wrong URL or non existing one or failure of requests.get
        with patch("weaviate.util.requests") as mock_obj:
            result_mock = Mock()
            result_mock.status_code = 404
            mock_obj.get.return_value = result_mock
            with self.assertRaises(ValueError) as error:
                _get_dict_from_object("http://www.url.com")
            check_error_message(self, error, url_error_message + "http://www.url.com")
            mock_obj.get.assert_called()

        # valid calls
        self.assertEqual(_get_dict_from_object({"key": "val"}), {"key": "val"})
        # read from file
        path = "/".join(__file__.split("/")[:-1])
        self.assertEqual(
            _get_dict_from_object(f"{path}/schema/schema_company.json"), schema_company
        )
        # read from URL
        path = "https://raw.githubusercontent.com/semi-technologies/weaviate-python-client/main/test/schema/schema_company.json"
        self.assertEqual(_get_dict_from_object(path), schema_company)

    def test_is_weaviate_object_url(self):
        """
        Test the `is_weaviate_object_url` function.
        """

        # valid formats
        self.assertTrue(
            is_weaviate_object_url("weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        )
        self.assertTrue(
            is_weaviate_object_url(
                "weaviate://some-domain.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"
            )
        )

        # invalid formats
        ## wrong argument data type
        self.assertFalse(
            is_weaviate_object_url(["weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"])
        )
        ## wrong prefix, i.e. does not start with 'weaviate://'
        self.assertFalse(
            is_weaviate_object_url("http://some-domain.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        )
        ## wrong path, additional '/thing'
        self.assertFalse(
            is_weaviate_object_url("weaviate://localhost/things/f61b-b524-45e0-9bbe-2c1550bf73d2")
        )
        ## wrong domain format
        self.assertFalse(
            is_weaviate_object_url(
                "weaviate://some-INVALID-domain/28f3f61b-b524-45e0-9bbe-2c1550bf73d2"
            )
        )
        # wrong UUID format
        self.assertFalse(is_weaviate_object_url("weaviate://localhost/UUID"))

    def test_is_object_url(self):
        """
        Test the `is_object_url` function.
        """

        # valid formats
        self.assertTrue(
            is_object_url("http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        )
        self.assertTrue(
            is_object_url("http://ramalamadingdong/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        )

        # invalid formats
        ## wrong path, should be at least 3 subpaths to the object UUID
        self.assertFalse(is_object_url("objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"))
        ## wrong '/v2', shoudl be '/v1'
        self.assertFalse(
            is_object_url("http://localhost:8080/v2/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        )
        ## wrong UUID format
        self.assertFalse(
            is_object_url("http://ramalamadingdong/v1/objects/1c9cd584-88fe-5010-83d0")
        )
        ## wrong objects path, instead of '/passions' should have been '/objects/
        self.assertFalse(
            is_object_url("http://localhost:8080/v1/passions/1c9cd584-88fe-5010-83d0-017cb3fcb446")
        )

    def test_get_valid_uuid(self):
        """
        Test the `get_valid_uuid` function.
        """

        # valid calls
        result = get_valid_uuid("weaviate://localhost/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        self.assertEqual(result, "28f3f61b-b524-45e0-9bbe-2c1550bf73d2")

        result = get_valid_uuid("weaviate://otherhost.com/28f3f61b-b524-45e0-9bbe-2c1550bf73d2")
        self.assertEqual(result, "28f3f61b-b524-45e0-9bbe-2c1550bf73d2")

        result = get_valid_uuid(
            "http://localhost:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"
        )
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid(
            "http://otherhost_2:8080/v1/objects/1c9cd584-88fe-5010-83d0-017cb3fcb446"
        )
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid(
            "http://otherhost_2:8080/v1/objects/1c9cd58488fe501083d0017cb3fcb446"
        )
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid("1c9cd584-88fe-5010-83d0-017cb3fcb446")
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid("1c9cd58488fe501083d0017cb3fcb446")
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        result = get_valid_uuid(uuid_lib.UUID("1c9cd58488fe501083d0017cb3fcb446"))
        self.assertEqual(result, "1c9cd584-88fe-5010-83d0-017cb3fcb446")

        # invalid formats
        type_error_message = "'uuid' must be of type str or uuid.UUID, but was: "
        value_error_message = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        ## neither an object URL nor a weaviate object URL
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v1/1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        # wrong UUID format
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v1/objects/some-UUID")
        check_error_message(self, error, value_error_message)

        ## wrong '/v2', shoudl be '/v1'
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("http://localhost:8080/v2/objects/1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        ## wrong URL
        with self.assertRaises(ValueError) as error:
            get_valid_uuid("weaviate://INVALID_URL//1c9cd584-88fe-5010-83d0-017cb3fcb")
        check_error_message(self, error, value_error_message)

        ## wrong UUID data type
        with self.assertRaises(TypeError) as error:
            get_valid_uuid(12)
        check_error_message(self, error, type_error_message + str(int))

    def test_get_vector(self):
        """
        Test the `get_vector` function.
        """

        vector_list = [1.0, 2.0, 3.0]
        # vector is a list
        self.assertEqual(get_vector(vector_list), vector_list)

        # vector is a `torch.Tensor` or `numpy.ndarray`
        vector_mock = Mock()
        mock_tolist = Mock()
        mock_tolist.tolist.return_value = vector_list
        mock_squeeze = Mock(return_value=mock_tolist)
        vector_mock.squeeze = mock_squeeze
        self.assertEqual(get_vector(vector_mock), vector_list)

        # vector is a `tf.Tensor`
        vector_mock = Mock()
        mock_tolist = Mock()
        mock_squeeze = Mock()
        mock_tolist.tolist.return_value = vector_list
        mock_squeeze.squeeze.return_value = mock_tolist
        mock_numpy = Mock(return_value=mock_squeeze)
        vector_mock.numpy = mock_numpy
        vector_mock.squeeze = Mock(side_effect=AttributeError("TEST TensorFlow Tensor"))
        self.assertEqual(get_vector(vector_mock), vector_list)

        # invalid call
        type_error_message = (
            "The type of the 'vector' argument is not supported!\n"
            "Supported types are `list`, 'numpy.ndarray`, `torch.Tensor` "
            "and `tf.Tensor`"
        )
        with self.assertRaises(TypeError) as error:
            get_vector("[1., 2., 3.]")
        check_error_message(self, error, type_error_message)

    def test_get_domain_from_weaviate_url(self):
        """
        Test the `get_domain_from_weaviate_url` function.
        """

        uuid = "28f3f61b-b524-45e0-9bbe-2c1550bf73d2"
        self.assertEqual(get_domain_from_weaviate_url(f"weaviate://localhost/{uuid}"), "localhost")
        self.assertEqual(get_domain_from_weaviate_url(f"weaviate://otherhost/{uuid}"), "otherhost")

    def test__is_sub_schema(self):
        """
        Test the `_is_sub_schema` function.
        """

        self.assertTrue(_is_sub_schema(schema_set, schema_set))
        self.assertTrue(_is_sub_schema(schema_set["classes"][0], schema_set))
        self.assertTrue(_is_sub_schema(schema_sub_set, schema_set))

        schema_set_copy = deepcopy(schema_set)
        for schema_class in schema_set_copy["classes"]:
            schema_class["class"] = schema_class["class"].lower()
        self.assertTrue(_is_sub_schema(schema_set_copy, schema_set))
        self.assertTrue(_is_sub_schema(schema_set, schema_set_copy))
        self.assertTrue(_is_sub_schema(schema_set_copy, schema_set_copy))

        self.assertFalse(_is_sub_schema({"class": "A"}, schema_set))
        self.assertFalse(_is_sub_schema(disjoint_set, schema_set))
        self.assertFalse(_is_sub_schema(partial_set, schema_set))
        self.assertFalse(_is_sub_schema(schema_set_extended_prop, schema_set))

        # invalid calls

        invalid_sub_schema_msg = "The sub schema class/es MUST have a 'class' keyword each!"

        with self.assertRaises(SchemaValidationException) as error:
            _is_sub_schema({}, schema_set)
        check_error_message(self, error, invalid_sub_schema_msg)

    def test_image_encoder_b64(self):
        """
        Test the `image_encoder_b64` function.
        """

        encode_weaviate_logo_b64 = " "
        file_path_value_error_message = "No file found at location non-existingfile.png"
        type_error_message = (
            '"image_or_image_path" should be a image path or a binary read file'
            " (io.BufferedReader)"
        )

        # invalid calls
        with self.assertRaises(ValueError) as error:
            image_encoder_b64("non-existingfile.png")
        check_error_message(self, error, file_path_value_error_message)

        with self.assertRaises(TypeError) as error:
            image_encoder_b64(True)
        check_error_message(self, error, type_error_message)

        with self.assertRaises(TypeError) as error:
            with open("image.png", "wb") as file:
                image_encoder_b64(file)
        check_error_message(self, error, type_error_message)

        # valid calls
        encrypted_1 = image_encoder_b64("integration/weaviate-logo.png")
        self.assertEqual(encrypted_1, encode_weaviate_logo_b64)
        self.assertIsInstance(encrypted_1, str)

        with open("integration/weaviate-logo.png", "rb") as file:
            encrypted_2 = image_encoder_b64(file)
        self.assertEqual(encrypted_2, encode_weaviate_logo_b64)
        self.assertIsInstance(encrypted_2, str)

    def test_image_decoder_b64(self):
        """
        Test the `image_decoder_b64` function.
        """

        encode_weaviate_logo_b64 = " "

        decoded = image_decoder_b64(encode_weaviate_logo_b64)

        with open("integration/weaviate-logo.png", "rb") as file:
            self.assertEqual(decoded, file.read())
            self.assertIsInstance(decoded, bytes)

    @patch("weaviate.util.uuid_lib")
    def test_generate_uuid5(self, mock_uuid):

        result = generate_uuid5("TestID!", "Test!")
        self.assertIsInstance(result, str)
        mock_uuid.uuid5.assert_called()



"""
GraphQL `Get` command.
"""
from dataclasses import dataclass
from json import dumps
from typing import List, Union, Optional, Dict, Tuple

from weaviate.connect import Connection
from weaviate.gql.filter import (
    Where,
    NearText,
    NearVector,
    GraphQL,
    NearObject,
    Filter,
    Ask,
    NearImage,
    Sort,
)
from weaviate.types import UUID
from weaviate.util import image_encoder_b64, _capitalize_first_letter, get_valid_uuid
from weaviate.warnings import _Warnings


@dataclass
class BM25:
    query: str
    properties: Optional[List[str]]

    def __str__(self) -> str:
        ret = f'query: "{self.query}"'
        if self.properties is not None and len(self.properties) > 0:
            props = '","'.join(self.properties)
            ret += f', properties: ["{props}"]'
        return "bm25:{" + ret + "}"


@dataclass
class Hybrid:
    query: str
    alpha: float
    vector: List[float]

    def __str__(self) -> str:
        ret = f'query: "{self.query}"'
        if self.vector is not None:
            ret += f", vector: {self.vector}"
        if self.alpha is not None:
            ret += f", alpha: {self.alpha}"

        return "hybrid:{" + ret + "}"


class GetBuilder(GraphQL):
    """
    GetBuilder class used to create GraphQL queries.
    """

    def __init__(
        self, class_name: str, properties: Union[List[str], str, None], connection: Connection
    ):
        """
        Initialize a GetBuilder class instance.

        Parameters
        ----------
        class_name : str
            Class name of the objects to interact with.
        properties : str or list of str
            Properties of the objects to interact with.
        connection : weaviate.connect.Connection
            Connection object to an active and running Weaviate instance.

        Raises
        ------
        TypeError
            If argument/s is/are of wrong type.
        """

        super().__init__(connection)

        if not isinstance(class_name, str):
            raise TypeError(f"class name must be of type str but was {type(class_name)}")
        if properties is None:
            properties = []
        if isinstance(properties, str):
            properties = [properties]
        if not isinstance(properties, list):
            raise TypeError(
                "properties must be of type str, " f"list of str or None but was {type(properties)}"
            )
        for prop in properties:
            if not isinstance(prop, str):
                raise TypeError("All the `properties` must be of type `str`!")

        self._class_name: str = _capitalize_first_letter(class_name)
        self._properties: List[str] = properties
        self._additional: dict = {"__one_level": set()}
        # '__one_level' refers to the additional properties that are just a single word, not a dict
        # thus '__one_level', only one level of complexity
        self._where: Optional[Where] = None  # To store the where filter if it is added
        self._limit: Optional[str] = None  # To store the limit filter if it is added
        self._offset: Optional[str] = None  # To store the offset filter if it is added
        self._after: Optional[str] = None  # To store the offset filter if it is added
        self._near_ask: Optional[Filter] = None  # To store the `near`/`ask` clause if it is added
        self._contains_filter = False  # true if any filter is added
        self._sort: Optional[Sort] = None
        self._bm25: Optional[BM25] = None
        self._hybrid: Optional[Hybrid] = None
        self._alias: Optional[str] = None

    def with_after(self, after_uuid: UUID):
        """Can be used to extract all elements by giving the last ID from the previous "page".

        Requires limit to be set but cannot be combined with any other filters or search. Part of the Cursor API."""
        if not isinstance(after_uuid, UUID.__args__):  # __args__ is workaround for python 3.8
            raise TypeError("after_uuid must be of type UUID (str or uuid.UUID)")

        self._after = f'after: "{get_valid_uuid(after_uuid)}"'
        self._contains_filter = True
        return self

    def with_where(self, content: dict) -> "GetBuilder":
        """
        Set `where` filter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter to set. See examples below.

        Examples
        --------
        The `content` prototype is like this:

        >>> content = {
        ...     'operator': '<operator>',
        ...     'operands': [
        ...         {
        ...             'path': [path],
        ...             'operator': '<operator>'
        ...             '<valueType>': <value>
        ...         },
        ...         {
        ...             'path': [<matchPath>],
        ...             'operator': '<operator>',
        ...             '<valueType>': <value>
        ...         }
        ...     ]
        ... }

        This is a complete `where` filter but it does not have to be like this all the time.

        Single operand:

        >>> content = {
        ...     'path': ["wordCount"],    # Path to the property that should be used
        ...     'operator': 'GreaterThan',  # operator
        ...     'valueInt': 1000       # value (which is always = to the type of the path property)
        ... }

        Or

        >>> content = {
        ...     'path': ["id"],
        ...     'operator': 'Equal',
        ...     'valueString': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf"
        ... }

        Multiple operands:

        >>> content = {
        ...     'operator': 'And',
        ...     'operands': [
        ...         {
        ...             'path': ["wordCount"],
        ...             'operator': 'GreaterThan',
        ...             'valueInt': 1000
        ...         },
        ...         {
        ...             'path': ["wordCount"],
        ...             'operator': 'LessThan',
        ...             'valueInt': 1500
        ...         }
        ...     ]
        ... }

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.
        """

        self._where = Where(content)
        self._contains_filter = True
        return self

    @property
    def name(self) -> str:
        return self._alias if self._alias else self._class_name

    def with_near_text(self, content: dict) -> "GetBuilder":
        """
        Set `nearText` filter. This filter can be used with text modules (text2vec).
        E.g.: text2vec-contextionary, text2vec-transformers.
        NOTE: The 'autocorrect' field is enabled only with the `text-spellcheck` Weaviate module.

        Parameters
        ----------
        content : dict
            The content of the `nearText` filter to set. See examples below.

        Examples
        --------
        Content full prototype:

        >>> content = {
        ...     'concepts': <list of str or str>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'moveAwayFrom': { # Optional
        ...         'concepts': <list of str or str>,
        ...         'force': <float>
        ...     },
        ...     'moveTo': { # Optional
        ...         'concepts': <list of str or str>,
        ...         'force': <float>
        ...     },
        ...     'autocorrect': <bool>, # Optional
        ... }

        Full content:

        >>> content = {
        ...     'concepts': ["fashion"],
        ...     'certainty': 0.7, # or 'distance'
        ...     'moveAwayFrom': {
        ...         'concepts': ["finance"],
        ...         'force': 0.45
        ...     },
        ...     'moveTo': {
        ...         'concepts': ["haute couture"],
        ...         'force': 0.85
        ...     },
        ...     'autocorrect': True
        ... }

        Partial content:

        >>> content = {
        ...     'concepts': ["fashion"],
        ...     'certainty': 0.7, # or 'distance'
        ...     'moveTo': {
        ...         'concepts': ["haute couture"],
        ...         'force': 0.85
        ...     }
        ... }

        Minimal content:

        >>> content = {
        ...     'concepts': "fashion"
        ... }

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        if self._near_ask is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        self._near_ask = NearText(content)
        self._contains_filter = True
        return self

    def with_near_vector(self, content: dict) -> "GetBuilder":
        """
        Set `nearVector` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearVector` filter to set. See examples below.

        Examples
        --------
        Content full prototype:

        >>> content = {
        ...     'vector' : <list of float>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        NOTE: Supported types for 'vector' are `list`, 'numpy.ndarray`, `torch.Tensor`
                and `tf.Tensor`.

        Full content:

        >>> content = {
        ...     'vector' : [.1, .2, .3, .5],
        ...     'certainty': 0.75, # or 'distance'
        ... }

        Minimal content:

        >>> content = {
        ...     'vector' : [.1, .2, .3, .5]
        ... }

        Or

        >>> content = {
        ...     'vector' : torch.tensor([.1, .2, .3, .5])
        ... }

        Or

        >>> content = {
        ...     'vector' : torch.tensor([[.1, .2, .3, .5]]) # it is going to be squeezed.
        ... }

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        if self._near_ask is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        self._near_ask = NearVector(content)
        self._contains_filter = True
        return self

    def with_near_object(self, content: dict) -> "GetBuilder":
        """
        Set `nearObject` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearObject` filter to set. See examples below.

        Examples
        --------
        Content prototype:

        >>> {
        ...     'id': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }
        >>> # alternatively
        >>> {
        ...     'beacon': "weaviate://localhost/ClassName/e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf"
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        is_server_version_14 = self._connection.server_version >= "1.14"

        if self._near_ask is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        self._near_ask = NearObject(content, is_server_version_14)
        self._contains_filter = True
        return self

    def with_near_image(self, content: dict, encode: bool = True) -> "GetBuilder":
        """
        Set `nearImage` filter.

        Parameters
        ----------
        content : dict
            The content of the `nearObject` filter to set. See examples below.
        encode : bool, optional
            Whether to encode the `content["image"]` to base64 and convert to string. If True, the
            `content["image"]` can be an image path or a file opened in binary read mode. If False,
            the `content["image"]` MUST be a base64 encoded string (NOT bytes, i.e. NOT binary
            string that looks like this: b'BASE64ENCODED' but simple 'BASE64ENCODED').
            By default True.

        Examples
        --------
        Content prototype:

        >>> content = {
        ...     'image': <str or binary read file>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ... }

        >>> {
        ...     'image': "e5dc4a4c-ef0f-3aed-89a3-a73435c6bbcf",
        ...     'certainty': 0.7 # or 'distance'
        ... }

        With `encoded` True:

        >>> content = {
        ...     'image': "my_image_path.png",
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Image', 'description')\\
        ...     .with_near_image(content, encode=True) # <- encode MUST be set to True

        OR

        >>> my_image_file = open("my_image_path.png", "br")
        >>> content = {
        ...     'image': my_image_file,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Image', 'description')\\
        ...     .with_near_image(content, encode=True) # <- encode MUST be set to True
        >>> my_image_file.close()

        With `encoded` False:

        >>> from weaviate.util import image_encoder_b64, image_decoder_b64
        >>> encoded_image = image_encoder_b64("my_image_path.png")
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Image', 'description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        OR

        >>> from weaviate.util import image_encoder_b64, image_decoder_b64
        >>> with open("my_image_path.png", "br") as my_image_file:
        ...     encoded_image = image_encoder_b64(my_image_file)
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Image', 'description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        Encode Image yourself:

        >>> import base64
        >>> with open("my_image_path.png", "br") as my_image_file:
        ...     encoded_image = base64.b64encode(my_image_file.read()).decode("utf-8")
        >>> content = {
        ...     'image': encoded_image,
        ...     'certainty': 0.7 # or 'distance' instead
        ... }
        >>> query = client.query.get('Image', 'description')\\
        ...     .with_near_image(content, encode=False) # <- encode MUST be set to False

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        AttributeError
            If another 'near' filter was already set.
        """

        if self._near_ask is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        if encode:
            content["image"] = image_encoder_b64(content["image"])
        self._near_ask = NearImage(content)
        self._contains_filter = True
        return self

    def with_limit(self, limit: int) -> "GetBuilder":
        """
        The limit of objects returned.

        Parameters
        ----------
        limit : int
            The max number of objects returned.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        ValueError
            If 'limit' is non-positive.
        """

        if limit < 1:
            raise ValueError("limit cannot be non-positive (limit >=1).")

        self._limit = f"limit: {limit} "
        self._contains_filter = True
        return self

    def with_offset(self, offset: int) -> "GetBuilder":
        """
        The offset of objects returned, i.e. the starting index of the returned objects should be
        used in conjunction with the `with_limit` method.

        Parameters
        ----------
        offset : int
            The offset used for the returned objects.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        ValueError
            If 'offset' is non-positive.
        """

        if offset < 0:
            raise ValueError("offset cannot be non-positive (offset >=0).")

        self._offset = f"offset: {offset} "
        self._contains_filter = True
        return self

    def with_ask(self, content: dict) -> "GetBuilder":
        """
        Ask a question for which weaviate will retrieve the answer from your data.
        This filter can be used only with QnA module: qna-transformers.
        NOTE: The 'autocorrect' field is enabled only with the `text-spellcheck` Weaviate module.

        Parameters
        ----------
        content : dict
            The content of the `ask` filter to set. See examples below.

        Examples
        --------
        Content full prototype:

        >>> content = {
        ...     'question' : <str>,
        ...     # certainty ONLY with `cosine` distance specified in the schema
        ...     'certainty': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'distance': <float>, # Optional, either 'certainty' OR 'distance'
        ...     'properties': <list of str or str> # Optional
        ...     'autocorrect': <bool>, # Optional
        ... }

        Full content:

        >>> content = {
        ...     'question' : "What is the NLP?",
        ...     'certainty': 0.7, # or 'distance'
        ...     'properties': ['body'] # search the answer in these properties only.
        ...     'autocorrect': True
        ... }

        Minimal content:

        >>> content = {
        ...     'question' : "What is the NLP?"
        ... }

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.
        """

        if self._near_ask is not None:
            raise AttributeError(
                "Cannot use multiple 'near' filters, or a 'near' filter along"
                " with a 'ask' filter!"
            )
        self._near_ask = Ask(content)
        self._contains_filter = True
        return self

    def with_additional(
        self, properties: Union[List, str, Dict[str, Union[List[str], str]], Tuple[dict, dict]]
    ) -> "GetBuilder":
        """
        Add additional properties (i.e. properties from `_additional` clause). See Examples below.
        If the the 'properties' is of data type `str` or `list` of `str` then the method is
        idempotent, if it is of type `dict` or `tuple` then the exiting property is going to be
        replaced. To set the setting of one of the additional property use the `tuple` data type
        where `properties` look like this (clause: dict, settings: dict) where the 'settings' are
        the properties inside the '(...)' of the clause. See Examples for more information.

        Parameters
        ----------
        properties : str, list of str, dict[str, str], dict[str, list of str] or tuple[dict, dict]
            The additional properties to include in the query. Can be property name as `str`,
            a list of property names, a dictionary (clause without settings) where the value is a
            `str` or list of `str`, or a `tuple` of 2 elements:
                (clause: Dict[str, str or list[str]], settings: Dict[str, Any])
            where the 'clause' is the property and all its sub-properties and the 'settings' is the
            setting of the property, i.e. everything that is inside the `(...)` right after the
            property name. See Examples below.

        Examples
        --------

        >>> # single additional property with this GraphQL query
        >>> '''
        ... {
        ...     Get {
        ...         Article {
        ...             title
        ...             author
        ...             _additional {
        ...                 id
        ...             }
        ...         }
        ...     }
        ... }
        ... '''
        >>> client.query\\
        ...     .get('Article', ['title', 'author'])\\
        ...     .with_additional('id']) # argument as `str`

        >>> # multiple additional property with this GraphQL query
        >>> '''
        ... {
        ...     Get {
        ...         Article {
        ...             title
        ...             author
        ...             _additional {
        ...                 id
        ...                 certainty
        ...             }
        ...         }
        ...     }
        ... }
        ... '''
        >>> client.query\\
        ...     .get('Article', ['title', 'author'])\\
        ...     .with_additional(['id', 'certainty']) # argument as `List[str]`

        >>> # additional properties as clause with this GraphQL query
        >>> '''
        ... {
        ...     Get {
        ...         Article {
        ...             title
        ...             author
        ...             _additional {
        ...                 classification {
        ...                     basedOn
        ...                     classifiedFields
        ...                     completed
        ...                     id
        ...                     scope
        ...                 }
        ...             }
        ...         }
        ...     }
        ... }
        ... '''
        >>> client.query\\
        ...     .get('Article', ['title', 'author'])\\
        ...     .with_additional(
        ...         {
        ...             'classification' : ['basedOn', 'classifiedFields', 'completed', 'id']
        ...         }
        ...     ) # argument as `dict[str, List[str]]`
        >>> # or with this GraphQL query
        >>> '''
        ... {
        ...     Get {
        ...         Article {
        ...             title
        ...             author
        ...             _additional {
        ...                 classification {
        ...                     completed
        ...                 }
        ...             }
        ...         }
        ...     }
        ... }
        ... '''
        >>> client.query\\
        ...     .get('Article', ['title', 'author'])\\
        ...     .with_additional(
        ...         {
        ...             'classification' : 'completed'
        ...         }
        ...     ) # argument as `Dict[str, str]`

        Consider the following GraphQL clause:

        >>> '''
        ... {
        ...     Get {
        ...         Article {
        ...             title
        ...             author
        ...             _additional {
        ...                 token (
        ...                     properties: ["content"]
        ...                     limit: 10
        ...                     certainty: 0.8
        ...                 ) {
        ...                     certainty
        ...                     endPosition
        ...                     entity
        ...                     property
        ...                     startPosition
        ...                     word
        ...                 }
        ...             }
        ...         }
        ...     }
        ... }
        ... '''

        Then the python translation of this is the following:

        >>> clause = {
        ...     'token': [ # if only one, can be passes as `str`
        ...         'certainty',
        ...         'endPosition',
        ...         'entity',
        ...         'property',
        ...         'startPosition',
        ...         'word',
        ...     ]
        ... }
        >>> settings = {
        ...     'properties': ["content"],  # is required
        ...     'limit': 10,                # optional, int
        ...     'certainty': 0.8            # optional, float
        ... }
        >>> client.query\\
        ...     .get('Article', ['title', 'author'])\\
        ...     .with_additional(
        ...         (clause, settings)
        ...     ) # argument as `Tuple[Dict[str, List[str]], Dict[str, Any]]`

        If the desired clause does not match any example above, then the clause can always be
        converted to string before passing it to the `.with_additional()` method.

        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.

        Raises
        ------
        TypeError
            If one of the property is not of a correct data type.
        """

        if isinstance(properties, str):
            self._additional["__one_level"].add(properties)
            return self

        if isinstance(properties, list):
            for prop in properties:
                if not isinstance(prop, str):
                    raise TypeError(
                        "If type of 'properties' is `list` then all items must be of type `str`!"
                    )
                self._additional["__one_level"].add(prop)
            return self

        if isinstance(properties, tuple):
            self._tuple_to_dict(properties)
            return self

        if not isinstance(properties, dict):
            raise TypeError(
                "The 'properties' argument must be either of type `str`, `list`, `dict` or "
                f"`tuple`! Given: {type(properties)}"
            )

        # only `dict` type here
        for key, values in properties.items():
            if not isinstance(key, str):
                raise TypeError(
                    "If type of 'properties' is `dict` then all keys must be of type `str`!"
                )
            self._additional[key] = set()
            if isinstance(values, str):
                self._additional[key].add(values)
                continue
            if not isinstance(values, list):
                raise TypeError(
                    "If type of 'properties' is `dict` then all the values must be either of type "
                    f"`str` or `list` of `str`! Given: {type(values)}!"
                )
            if len(values) == 0:
                raise ValueError(
                    "If type of 'properties' is `dict` and a value is of type `list` then at least"
                    " one element should be present!"
                )
            for value in values:
                if not isinstance(value, str):
                    raise TypeError(
                        "If type of 'properties' is `dict` and a value is of type `list` then all "
                        "items must be of type `str`!"
                    )
                self._additional[key].add(value)
        return self

    def with_sort(self, content: Union[list, dict]) -> "GetBuilder":
        """
        Sort objects based on specific field/s. Multiple sort fields can be used, the objects are
        going to be sorted according to order of the sort configs passed. This method can be called
        multiple times and it does not overwrite the last entry but appends it to the previous
        ones, see examples below.

        Parameters
        ----------
        content : Union[list, dict]
            The content of the Sort filter. Can be a single Sort configuration or a list of
            configurations.

        Examples
        --------
        The `content` should have this form:

        >>> content = {
        ...     'path': ['name']       # Path to the property that should be used
        ...     'order': 'asc'         # Sort order, possible values: asc, desc
        ... }
        >>> client.query.get('Author', ['name', 'address'])\\
        ...     .with_sort(content)

        Or a list of sort configurations:

        >>> content = [
        ...     {
        ...         'path': ['name']        # Path to the property that should be used
        ...         'order': 'asc'          # Sort order, possible values: asc, desc
        ...     },
        ...         'path': ['address']     # Path to the property that should be used
        ...         'order': 'desc'         # Sort order, possible values: asc, desc
        ...     }
        ... ]

        If we have a list we can add it in 2 ways.
        Pass the list:

        >>> client.query.get('Author', ['name', 'address'])\\
        ...     .with_sort(content)

        Or one configuration at a time:

        >>> client.query.get('Author', ['name', 'address'])\\
        ...     .with_sort(content[0])
        ...     .with_sort(content[1])

        It is possible to call this method multiple times with lists only too.


        Returns
        -------
        weaviate.gql.get.GetBuilder
            The updated GetBuilder.
        """

        if self._sort is None:
            self._sort = Sort(content=content)
            self._contains_filter = True
        else:
            self._sort.add(content=content)
        return self

    def with_bm25(self, query: str, properties: Optional[List[str]] = None) -> "GetBuilder":
        """Add BM25 query to search the inverted index for keywords.

        Parameters
        ----------
        query: str
            The query to search for.
        properties: Optional[List[str]]
            Which properties should be searched. If 'None' or empty all properties will be searched. By default, None
        """
        self._bm25 = BM25(query, properties)
        self._contains_filter = True

        return self

    def with_hybrid(
        self, query: str, alpha: Optional[float] = None, vector: Optional[List[float]] = None
    ):
        """Get objects using bm25 and vector, then combine the results using a reciprocal ranking algorithm.

        Parameters
        ----------
        query: str
            The query to search for.
        alpha: Optional[float]
            Factor determining how BM25 and vector search are weighted. If 'None' the weaviate default of 0.75 is used.
            By default, None
            alpha = 0 -> bm25, alpha=1 -> vector search
        vector: Optional[List[float]]
            Vector that is searched for. If 'None', weaviate will use the configured text-to-vector module to create a
            vector from the "query" field.
            By default, None
        """
        self._hybrid = Hybrid(query, alpha, vector)
        self._contains_filter = True
        return self

    def with_generate(
        self, single_prompt: Optional[str] = None, grouped_task: Optional[str] = None
    ) -> "GetBuilder":
        """Generate responses using the OpenAI generative search.

        At least one of the two parameters must be set.

        Parameters
        ----------
        grouped_task: Optional[str]
            The task to generate a grouped response. One
        single_prompt: Optional[str]
            The prompt to generate a single response.
        """
        if single_prompt is None and grouped_task is None:
            raise TypeError(
                "Either parameter grouped_result_task or single_result_prompt must be not None."
            )
        if (single_prompt is not None and not isinstance(single_prompt, str)) or (
            grouped_task is not None and not isinstance(grouped_task, str)
        ):
            raise TypeError("prompts and tasks must be of type str.")

        if self._connection.server_version < "1.17.3":
            _Warnings.weaviate_too_old_for_openai(self._connection.server_version)

        results: List[str] = ["error"]
        task_and_prompt = ""
        if single_prompt is not None:
            results.append("singleResult")
            task_and_prompt += f'singleResult:{{prompt:"{single_prompt}"}}'
        if grouped_task is not None:
            results.append("groupedResult")
            task_and_prompt += f'groupedResult:{{task:"{grouped_task}"}}'

        self._additional["__one_level"].add(f'generate({task_and_prompt}){{{" ".join(results)}}}')

        return self

    def with_alias(
        self,
        alias: str,
    ):
        """Gives an alias for the query. Needs to be used if 'multi_get' requests the same 'class_name' twice.

        Parameters
        ----------
        alias: str
            The alias for the query.
        """

        self._alias = alias
        return self

    def build(self, wrap_get: bool = True) -> str:
        """
        Build query filter as a string.

        Parameters
        ----------
        wrap_get: bool
            A boolean to decide wether {Get{...}} is placed around the query. Useful for multi_get.

        Returns
        -------
        str
            The GraphQL query as a string.
        """
        if wrap_get:
            query = "{Get{"
        else:
            query = ""

        if self._alias is not None:
            query += self._alias + ": "
        query += self._class_name
        if self._contains_filter:
            query += "("
            if self._where is not None:
                query += str(self._where)
            if self._limit is not None:
                query += self._limit
            if self._offset is not None:
                query += self._offset
            if self._near_ask is not None:
                query += str(self._near_ask)
            if self._sort is not None:
                query += str(self._sort)
            if self._bm25 is not None:
                query += str(self._bm25)
            if self._hybrid is not None:
                query += str(self._hybrid)
            if self._after is not None:
                query += self._after

            query += ")"

        additional_props = self._additional_to_str()

        if not (additional_props or self._properties):
            raise AttributeError(
                "No 'properties' or 'additional properties' specified to be returned. "
                "At least one should be included."
            )

        properties = " ".join(self._properties) + self._additional_to_str()
        query += "{" + properties + "}"
        if wrap_get:
            query += "}}"
        return query

    def _additional_to_str(self) -> str:
        """
        Convert `self._additional` attribute to a `str`.

        Returns
        -------
        str
            The converted self._additional.
        """

        str_to_return = " _additional {"

        has_values = False
        for one_level in sorted(self._additional["__one_level"]):
            has_values = True
            str_to_return += one_level + " "

        for key, values in sorted(self._additional.items(), key=lambda key_value: key_value[0]):
            if key == "__one_level":
                continue
            has_values = True
            str_to_return += key + " {"
            for value in sorted(values):
                str_to_return += value + " "
            str_to_return += "} "

        if has_values is False:
            return ""
        return str_to_return + "}"

    def _tuple_to_dict(self, tuple_value: tuple) -> None:
        """
        Convert the tuple data type argument to a dictionary.

        Parameters
        ----------
        tuple_value : tuple
            The tuple value as (clause: <dict>, settings: <dict>).

        Raises
        ------
        ValueError
            If 'tuple_value' does not have exactly 2 elements.
        TypeError
            If the configuration of the 'tuple_value' is not correct.
        """

        if len(tuple_value) != 2:
            raise ValueError(
                "If type of 'properties' is `tuple` then it should have length 2: "
                "(clause: <dict>, settings: <dict>)"
            )

        clause, settings = tuple_value
        if not isinstance(clause, dict) or not isinstance(settings, dict):
            raise TypeError(
                "If type of 'properties' is `tuple` then it should have this data type: "
                "(<dict>, <dict>)"
            )
        if len(clause) != 1:
            raise ValueError(
                "If type of 'properties' is `tuple` then the 'clause' (first element) should "
                f"have only one key. Given: {len(clause)}"
            )
        if len(settings) == 0:
            raise ValueError(
                "If type of 'properties' is `tuple` then the 'settings' (second element) should "
                f"have at least one key. Given: {len(settings)}"
            )

        clause_key, values = list(clause.items())[0]

        if not isinstance(clause_key, str):
            raise TypeError(
                "If type of 'properties' is `tuple` then first element's key should be of type "
                "`str`!"
            )

        clause_with_settings = clause_key + "("
        try:
            for key, value in sorted(settings.items(), key=lambda key_value: key_value[0]):
                if not isinstance(key, str):
                    raise TypeError(
                        "If type of 'properties' is `tuple` then the second elements (<dict>) "
                        "should have all the keys of type `str`!"
                    )
                clause_with_settings += key + ": " + dumps(value) + " "
        except TypeError:
            raise TypeError(
                "If type of 'properties' is `tuple` then the second elements (<dict>) "
                "should have all the keys of type `str`!"
            ) from None
        clause_with_settings += ")"

        self._additional[clause_with_settings] = set()
        if isinstance(values, str):
            self._additional[clause_with_settings].add(values)
            return
        if not isinstance(values, list):
            raise TypeError(
                "If type of 'properties' is `tuple` then first element's dict values must be "
                f"either of type `str` or `list` of `str`! Given: {type(values)}!"
            )
        if len(values) == 0:
            raise ValueError(
                "If type of 'properties' is `tuple` and first element's dict value is of type "
                "`list` then at least one element should be present!"
            )
        for value in values:
            if not isinstance(value, str):
                raise TypeError(
                    "If type of 'properties' is `tuple` and first element's dict value is of type "
                    " `list` then all items must be of type `str`!"
                )
            self._additional[clause_with_settings].add(value)



"""
WCS class definition.
"""
import time
from numbers import Real
from typing import Optional, List, Union, Dict, Tuple

from requests.exceptions import ConnectionError as RequestsConnectionError
from tqdm.auto import tqdm

from weaviate.auth import AuthClientPassword
from weaviate.connect.connection import BaseConnection
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
)


class WCS(BaseConnection):
    """
    WCS class used to create/delete WCS cluster instances.

    Attributes
    ----------
    dev : bool
        True if the WCS instance is for the development console, False if it is for the production
        environment.
    """

    # def __getattr__(self, attr_name):
    #     if attr_name in ('delete', 'get', 'post', 'put', 'patch'):
    #         raise AttributeError
    #     if attr_name in ('_delete', '_get', '_post', '_put', '_patch'):
    #         attr_name = attr_name[1:]
    #     return super().__getattr__(attr_name)

    # def __getattribute__(self, attr_name):
    #     if attr_name in ('delete', 'get', 'post', 'put', 'patch'):
    #         raise AttributeError
    #     if attr_name in ('_delete', '_get', '_post', '_put', '_patch'):
    #         attr_name = attr_name[1:]
    #     return super().__getattribute__(attr_name)

    def __init__(
        self,
        auth_client_secret: AuthClientPassword,
        timeout_config: Union[Tuple[Real, Real], Real] = (2, 20),
        proxies: Union[dict, str, None] = None,
        trust_env: bool = False,
        dev: bool = False,
    ):
        """
        Initialize a WCS class instance.

        Parameters
        ----------
        auth_client_secret : AuthClientPassword
            Authentication credentials for the WCS.
        dev : bool, optional
            Whether to use the development environment, i.e. https://dev.console.semi.technology/.
            If False uses the production environment, i.e. https://console.semi.technology/.
            By default False.
        timeout_config : tuple(Real, Real) or Real, optional
            Set the timeout configuration for all requests to the Weaviate server. It can be a
            real number or, a tuple of two real numbers: (connect timeout, read timeout).
            If only one real number is passed then both connect and read timeout will be set to
            that value, by default (2, 20).
        """

        self.dev = dev

        if dev:
            url = "https://dev.wcs.api.semi.technology"
        else:
            url = "https://wcs.api.semi.technology"

        super().__init__(
            url=url,
            auth_client_secret=auth_client_secret,
            timeout_config=timeout_config,
            proxies=proxies,
            trust_env=trust_env,
            additional_headers=None,
            startup_period=None,
        )
        self._email = auth_client_secret.username

    def __del__(self):
        # in case an exception happens before definition of these members
        if hasattr(self, "_connection"):
            self._connection.close()

    def create(
        self,
        cluster_name: str = None,
        cluster_type: str = "sandbox",
        with_auth: bool = False,
        modules: Optional[Union[str, dict, list]] = None,
        config: dict = None,
        wait_for_completion: bool = True,
    ) -> str:
        """
        Create the cluster and return The Weaviate server URL.

        Parameters
        ----------
        cluster_name : str, optional
            The name of the weaviate cluster to be created, if None a random one is going to be
            generated, by default None.
            NOTE: Case insensitive. The created cluster's name is always lowercased.
        cluster_type : str, optional
            the cluster type/tier, by default 'sandbox'.
        with_auth : bool, optional
            Enable the authentication to the cluster about to be created,
            by default False.
        modules: str or dict or list, optional
            The modules to use, can have multiple modules. One module should look like this:
            >>> {
            ...     "name": "string", # required
            ...     "repo": "string", # optional
            ...     "tag": "string", # optional
            ...     "inferenceUrl": "string" # optional
            ... }
            See the Examples for additional information.
        config : dict, optional
            the cluster configuration. If NOT None then `cluster_name`, `cluster_type`, `module`
            are ignored and the whole cluster configuration should be in this argument,
            by default None. See the Examples below for the complete configuration schema.
        wait_for_completion : bool, optional
            Whether to wait until the cluster is built,
            by default True

        Examples
        --------
        If the `modules` is string then it is going to be used as the MODULE_NAME with a default tag
        for that given MODULE_NAME. If `module` is a dict then it should have the below structure.

        Contextionary:

        >>> {
        ...     "name": "text2vec-contextionary",
        ...     "tag": "en0.16.0-v1.0.0" # this is the default tag
        ... }

        Transformers:

        >>> {
        ...     "name": "text2vec-transformers",
        ...     "tag": "sentence-transformers-paraphrase-MiniLM-L6-v2"
        ... }

        Both the examples above use the 'semitechnologies' repo (which is the default one).
        The `modules` also can be a list of individual module configuration that conforms to the
        above description.

        The COMPLETE `config` argument looks like this:

        >>> {
        ...     "email": "user@example.com",
        ...     "configuration": {
        ...         "requiresAuthentication": true,
        ...         "c11yTag": "string",
        ...         "tier": "string",
        ...         "supportLevel": "string",
        ...         "region": "string",
        ...         "release": {
        ...             "chart": "latest",
        ...             "weaviate": "latest"
        ...         },
        ...         "modules": [
        ...             {
        ...                 "name": "string",
        ...                 "repo": "string",
        ...                 "tag": "string",
        ...                 "inferenceUrl": "string"
        ...             }
        ...         ],
        ...         "backup": {
        ...             "activated": false
        ...         },
        ...         "restore": {
        ...             "name": "string"
        ...         }
        ...     },
        ...     "id": "string"
        ... }

        Returns
        -------
        str
            The URL of the create Weaviate server cluster.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If creating the weaviate cluster failed for a different reason,
            more information is given in the exception.
        TypeError
            If `config` is neither None nor of type 'dict'.
        TypeError
            If `modules` argument is of a wrong type.
        KeyError
            If one of the `modules` does not conform to the module schema.
        TypeError
            In case `modules` is a list and one module has a wrong type.
        TypeError
            In case one of the modules is of type dict and the values are not of type 'str'.
        """

        if cluster_name is not None:
            cluster_name = cluster_name.lower()

        if config is None:
            config = {
                "id": cluster_name,
                "configuration": {"tier": cluster_type, "requiresAuthentication": with_auth},
            }
            config["configuration"]["modules"] = _get_modules_config(modules)
        else:
            if not isinstance(config, dict):
                raise TypeError(
                    "The `config` argument must be either None or of type 'dict', given: "
                    f"{type(config)}"
                )
            if "id" in config:
                cluster_name = config["id"].lower()

        try:
            response = self.post(path="/clusters", weaviate_object=config)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("WCS cluster was not created.") from conn_err
        if response.status_code == 400 and "already exists" in response.text:
            # this line is never executed if cluster_name is None
            return "https://" + self.get_cluster_config(cluster_name)["meta"]["PublicURL"]

        if response.status_code != 202:
            raise UnexpectedStatusCodeException("Creating WCS instance", response)

        cluster_name = response.json()["id"]

        if wait_for_completion is True:
            title_bar = tqdm(
                bar_format="{desc}",
                desc="Creating cluster:",
                leave=True,
            )
            progress_bar = tqdm(
                total=100.0,
                leave=True,
                unit="%",
                bar_format=(
                    "{percentage:3.0f}% |{bar}|[{elapsed}<{remaining}, {rate_fmt}{postfix}]"
                ),
            )
            progress = 0
            while progress != 100:
                time.sleep(2.0)
                progress_state = self.get_cluster_config(cluster_name)["status"]["state"]
                progress = progress_state["percentage"]
                progress_bar.update(progress - progress_bar.n)
                title_bar.set_description(progress_state.get("message"))
            title_bar.close()
            progress_bar.close()

        return "https://" + self.get_cluster_config(cluster_name)["meta"]["PublicURL"]

    def is_ready(self, cluster_name: str) -> bool:
        """
        Check if the cluster is created.

        Parameters
        ----------
        cluster_name : str
            The name of the weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.

        Returns
        -------
        bool
            True if cluster is created and ready to use, False otherwise.
        """

        cluster_name = cluster_name.lower()
        response = self.get_cluster_config(cluster_name)
        if response == {}:
            raise ValueError(f"No cluster with name: '{cluster_name}'. Check the name again!")
        if response["status"]["state"]["percentage"] == 100:
            return True
        return False

    def get_clusters(self) -> Optional[List[str]]:
        """
        Lists all weaviate clusters registered with the this account.

        Returns
        -------
        Optional[List[str]]
            A list of cluster names or None if no clusters.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If getting the weaviate clusters failed for a different reason,
            more information is given in the exception.
        """

        try:
            response = self.get(
                path="/clusters/list",
                params={"email": self._email},
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("WCS clusters were not fetched.") from conn_err
        if response.status_code == 200:
            return response.json()["clusterIDs"]
        raise UnexpectedStatusCodeException("Checking WCS instance", response)

    def get_cluster_config(self, cluster_name: str) -> dict:
        """
        Get details of a cluster.

        Parameters
        ----------
        cluster_name : str
            The name of the weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.

        Returns
        -------
        dict
            Details in a JSON format. If no cluster with such name was found then an empty
            dictionary is returned.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If getting the weaviate cluster failed for a different reason,
            more information is given in the exception.
        """

        try:
            response = self.get(
                path="/clusters/" + cluster_name.lower(),
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("WCS cluster info was not fetched.") from conn_err
        if response.status_code == 200:
            return response.json()
        if response.status_code == 404:
            return {}
        raise UnexpectedStatusCodeException("Checking WCS instance", response)

    def delete_cluster(self, cluster_name: str) -> None:
        """
        Delete the WCS Weaviate cluster instance.

        Parameters
        ----------
        cluster_name : str
            The name of the weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If deleting the weaviate cluster failed for a different reason,
            more information is given in the exception.
        """

        try:
            response = self.delete(
                path="/clusters/" + cluster_name.lower(),
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("WCS cluster was not deleted.") from conn_err
        if response.status_code in (200, 404):
            return
        raise UnexpectedStatusCodeException("Deleting WCS instance", response)

    def get_users_of_cluster(self, cluster_name: str) -> list:
        """
        Get users of the WCS Weaviate cluster instance.

        Parameters
        ----------
        cluster_name : str
            The name of the Weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnsuccessfulStatusCodeError
            If deleting the Weaviate cluster failed for a different reason,
            more information is given in the exception.

        Returns
        -------
        list
            The list of users.
        """

        path = f"/clusters/{cluster_name.lower()}/users"

        try:
            response = self.get(
                path=path,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Could not get users of the cluster due to connection error."
            ) from conn_err
        if response.status_code == 200:
            return response.json()["users"]
        raise UnexpectedStatusCodeException("Getting cluster's users", response)

    def add_user_to_cluster(self, cluster_name: str, user: str) -> None:
        """
        Add user to the WCS Weaviate cluster instance.

        Parameters
        ----------
        cluster_name : str
            The name of the Weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.
        user:
            The user to be added to WCS Weaviate cluster instance.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnsuccessfulStatusCodeError
            If deleting the Weaviate cluster failed for a different reason,
            more information is given in the exception.
        """

        path = f"/clusters/{cluster_name.lower()}/users/{user}"

        try:
            response = self.post(
                path=path,
                weaviate_object=None,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Could not add user of the cluster due to connection error."
            ) from conn_err
        if response.status_code == 200:
            return
        raise UnexpectedStatusCodeException("Adding user to cluster", response)

    def remove_user_from_cluster(self, cluster_name: str, user: str) -> None:
        """
        Remove user from the WCS Weaviate cluster instance.

        Parameters
        ----------
        cluster_name : str
            The name of the Weaviate server cluster.
            NOTE: Case insensitive. The WCS cluster's name is always lowercased.
        user:
            The user to be removed from WCS Weaviate cluster instance.

        Raises
        ------
        requests.ConnectionError
            If the network connection to Weaviate fails.
        weaviate.UnsuccessfulStatusCodeError
            If deleting the Weaviate cluster failed for a different reason,
            more information is given in the exception.
        """

        path = f"/clusters/{cluster_name.lower()}/users/{user}"

        try:
            response = self.delete(
                path=path,
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                "Could not remove user from the cluster due to connection error."
            ) from conn_err
        if response.status_code == 200:
            return
        raise UnexpectedStatusCodeException("Removing user from cluster", response)


def _get_modules_config(modules: Optional[Union[str, dict, list]]) -> List[Dict[str, str]]:
    """
    Get an WCS modules configuration format.

    Parameters
    ----------
    modules : Optional[str, dict, list]
        The modules information from which to construct the modules configuration.

    Returns
    -------
    List[Dict[str, str]]
        The modules configuration as a list.

    Raises
    ------
    TypeError
        If `modules` argument is of a wrong type.
    KeyError
        If one of the `modules` does not conform to the module schema.
    TypeError
        In case `modules` is a list and one module has a wrong type.
    TypeError
        In case one of the modules is of type dict and the values are not of type 'str'.
    """

    def get_module_dict(module: Union[Dict[str, str], str]) -> Dict[str, str]:
        """
        Local function to validate each module configuration.

        Parameters
        ----------
        _module : Union[dict, str]
            The module configuration to be validated.

        Returns
        -------
        Dict[str, str]
            The configuration of the module as a dictionary.
        """

        if isinstance(module, str):
            # only module name
            return {"name": module}

        if isinstance(module, dict):
            # module config
            if "name" not in module or not set(module).issubset(
                ["name", "tag", "repo", "inferenceUrl"]
            ):
                raise KeyError(
                    "A module should have a required key: 'name',  and optional keys: 'tag', "
                    f"'repo' and/or 'inferenceUrl'! Given keys: {module.keys()}"
                )
            for key, value in module.items():
                if not isinstance(value, str):
                    raise TypeError(
                        "The type of each value of the module's dict should be 'str'! "
                        f"The key '{key}' has type: {type(value)}"
                    )
            return module

        raise TypeError(
            "Wrong type for one of the modules. Should be either 'str' or 'dict' but given: "
            f"{type(module)}"
        )

    if modules is None:
        # no module
        return []

    if isinstance(modules, (str, dict)):
        return [get_module_dict(modules)]
    if isinstance(modules, list):
        to_return = []
        for _module in modules:
            to_return.append(get_module_dict(_module))
        return to_return

    raise TypeError(
        "Wrong type for the `modules` argument. Accepted types are: NoneType, 'str', 'dict' or "
        f"`list` but given: {type(modules)}"
    )
